# -*- coding: utf-8 -*-

import numpy as np
from scipy import interpolate

from . import Filter  # prevent circular import in Python < 3.5


class Modulation(Filter):
    r"""Design a filter bank with a kernel centered at each frequency.

    Design a filter bank from translated versions of a mother filter.
    The mother filter is translated to each eigenvalue of the Laplacian via
    modulation. A signal is modulated by multiplying it with an eigenvector.
    Similarly to localization, it is an element-wise multiplication of a kernel
    with the columns of :attr:`pygsp2.graphs.Graph.U`, i.e., the eigenvectors,
    in the vertex domain.

    This filter bank can be used to compute the frequency content of a signal
    at each vertex. After filtering, one obtains a vertex-frequency
    representation :math:`Sf(i,k)` of a signal :math:`f` as

    .. math:: Sf(i, k) = \langle g_{i,k}, f \rangle,

    where :math:`g_{i,k}` is the mother kernel modulated in the spectral domain
    by the eigenvector :math:`u_k`, and localized on vertex :math:`v_i`.

    While :math:`g_{i,k}` should ideally be localized in both the spectral and
    vertex domains, that is impossible for some graphs due to the localization
    of some eigenvectors. See :attr:`pygsp2.graphs.Graph.coherence`.

    As modulation and localization don't commute, one can define the frame as
    :math:`g_{i,k} = T_i M_k g` (modulation first) or :math:`g_{i,k} = M_k T_i
    g` (localization first). Localization first usually gives better results.
    When localizing first, the obtained vertex-frequency representation is a
    generalization to graphs of the windowed graph Fourier transform. Indeed,

    .. math:: Sf(i, k) = \langle f^\text{win}_i, u_k \rangle

    is the graph Fourier transform of the windowed signal :math:`f^\text{win}`.
    The signal :math:`f` is windowed in the vertex domain by a point-wise
    multiplication with the localized kernel :math:`T_i g`.

    When localizing first, the spectral representation of the filter bank is
    different for every localization. As such, we always evaluate the filter in
    the spectral domain with modulation first. Moreover, the filter bank is
    only defined at the eigenvalues (as modulation is done with discrete
    eigenvectors). Evaluating it elsewhere returns NaNs.

    Parameters
    ----------
    graph : :class:`pygsp2.graphs.Graph`
    kernel : :class:`pygsp2.filters.Filter`
        Kernel function to be modulated.
    modulation_first : bool
        First modulate then localize the kernel if True, first localize then
        modulate if False. The two operators do not commute. This setting only
        applies to :meth:`filter`. :meth:`evaluate` only performs modulation,
        as the filter would otherwise have a different spectrum depending on
        where it is localized.

    See Also
    --------
    Gabor : Another way to translate a filter in the spectral domain.

    Notes
    -----
    The eigenvalues of the graph Laplacian (i.e., the Fourier basis) are needed
    to modulate the kernels.

    References
    ----------
    See :cite:`shuman2016vertexfrequency` for details on this vertex-frequency
    representation of graph signals.

    Examples
    --------

    Vertex-frequency representations.
    Modulating first doesn't produce sufficiently localized filters.

    >>> import matplotlib.pyplot as plt
    >>> G = graphs.Path(90)
    >>> G.compute_fourier_basis()
    >>>
    >>> # Design the filter banks.
    >>> g = filters.Heat(G, 500)
    >>> g1 = filters.Modulation(G, g, modulation_first=False)
    >>> g2 = filters.Modulation(G, g, modulation_first=True)
    >>> _ = g1.plot(sum=False, labels=False)
    >>>
    >>> # Signal.
    >>> s = np.empty(G.N)
    >>> s[:30] = G.U[:30, 10]
    >>> s[30:60] = G.U[30:60, 60]
    >>> s[60:] = G.U[60:, 30]
    >>> G.set_coordinates('line1D')
    >>> _ = G.plot(s)
    >>>
    >>> # Filter with both filter banks.
    >>> s1 = g1.filter(s)
    >>> s2 = g2.filter(s)
    >>>
    >>> # Visualize the vertex-frequency representation of the signal.
    >>> fig, axes = plt.subplots(1, 2)
    >>> _ = axes[0].imshow(np.abs(s1.T)**2)
    >>> _ = axes[1].imshow(np.abs(s2.T)**2)
    >>> _ = axes[0].set_title('localization then modulation')
    >>> _ = axes[1].set_title('modulation then localization')
    >>> ticks = [0, G.N//2, G.N-1]
    >>> labels = ['{:.1f}'.format(e) for e in G.e[ticks]]
    >>> _ = axes[0].set_yticks(ticks)
    >>> _ = axes[1].set_yticks([])
    >>> _ = axes[0].set_yticklabels(labels)
    >>> _ = axes[0].set_ylabel('graph frequency')
    >>> _ = axes[0].set_xlabel('node')
    >>> _ = axes[1].set_xlabel('node')
    >>> _ = axes[0].set_xticks(ticks)
    >>> _ = axes[1].set_xticks(ticks)
    >>> fig.tight_layout()
    >>>
    >>> # Reconstruction.
    >>> s = g2.filter(s2)
    >>> _ = G.plot(s)

    """

    def __init__(self, graph, kernel, modulation_first=False):

        self.G = graph
        self._kernels = kernel
        self._modulation_first = modulation_first

        if kernel.n_filters != 1:
            raise ValueError('A kernel must be one filter. The passed '
                             'filter bank {} has {}.'.format(
                                 kernel, kernel.n_filters))
        if kernel.G is not graph:
            raise ValueError('The graph passed to this filter bank must '
                             'be the one used to build the mother kernel.')

        self.n_features_in, self.n_features_out = (1, graph.n_vertices)
        self.n_filters = self.n_features_in * self.n_features_out
        self.Nf = self.n_filters  # TODO: kept for backward compatibility only.

    def evaluate(self, x):
        """TODO: will become _evaluate once polynomial filtering is merged."""

        if not hasattr(self, '_coefficients'):
            # Graph Fourier transform -> modulation -> inverse GFT.
            c = self.G.igft(self._kernels.evaluate(self.G.e).squeeze())
            c = np.sqrt(self.G.n_vertices) * self.G.U * c[:, np.newaxis]
            self._coefficients = self.G.gft(c)

        shape = x.shape
        x = x.flatten()
        y = np.full((self.n_features_out, x.size), np.nan)
        for i in range(len(x)):
            query = self._coefficients[x[i] == self.G.e]
            if len(query) != 0:
                y[:, i] = query[0]
        return y.reshape((self.n_features_out,) + shape)

    def filter(self, s, method='exact', order=None):
        """TODO: indirection will be removed when poly filtering is merged.
        TODO: with _filter and shape handled in Filter.filter, synthesis will work.
        """
        if self._modulation_first:
            return super(Modulation, self).filter(s, method='exact')
        else:
            # The dot product with each modulated kernel is equivalent to the
            # GFT, as for the localization and the IGFT.
            y = np.empty((self.G.n_vertices, self.G.n_vertices))
            for i in range(self.G.n_vertices):
                x = s * self._kernels.localize(i)
                y[i] = np.sqrt(self.G.n_vertices) * self.G.gft(x)
            return y



# -*- coding: utf-8 -*-

from __future__ import division

import numpy as np
from scipy import optimize

from pygsp2 import utils
from . import Filter  # prevent circular import in Python < 3.5


class Abspline(Filter):
    r"""Design an A B cubic spline wavelet filter bank.

    Parameters
    ----------
    G : graph
    Nf : int
        Number of filters from 0 to lmax (default = 6)
    lpfactor : float
        Low-pass factor lmin=lmax/lpfactor will be used to determine scales,
        the scaling function will be created to fill the lowpass gap.
        (default = 20)
    scales : ndarray
        Vector of scales to be used.
        By default, initialized with :func:`pygsp2.utils.compute_log_scales`.

    Examples
    --------

    Filter bank's representation in Fourier and time (ring graph) domains.

    >>> import matplotlib.pyplot as plt
    >>> G = graphs.Ring(N=20)
    >>> G.estimate_lmax()
    >>> G.set_coordinates('line1D')
    >>> g = filters.Abspline(G)
    >>> s = g.localize(G.N // 2)
    >>> fig, axes = plt.subplots(1, 2)
    >>> _ = g.plot(ax=axes[0])
    >>> _ = G.plot(s, ax=axes[1])

    """

    def __init__(self, G, Nf=6, lpfactor=20, scales=None):

        def kernel_abspline3(x, alpha, beta, t1, t2):
            M = np.array([[1, t1, t1**2, t1**3],
                          [1, t2, t2**2, t2**3],
                          [0, 1, 2*t1, 3*t1**2],
                          [0, 1, 2*t2, 3*t2**2]])
            v = np.array([1, 1, t1**(-alpha) * alpha * t1**(alpha - 1),
                          -beta*t2**(- beta - 1) * t2**beta])
            a = np.linalg.solve(M, v)

            r1 = x <= t1
            r2 = (x >= t1)*(x < t2)
            r3 = (x >= t2)

            if isinstance(x, np.float64):

                if r1:
                    r = x[r1]**alpha * t1**(-alpha)
                if r2:
                    r = a[0] + a[1] * x + a[2] * x**2 + a[3] * x**3
                if r3:
                    r = x[r3]**(-beta) * t2**beta

            else:
                r = np.zeros(x.shape)

                x2 = x[r2]

                r[r1] = x[r1]**alpha * t1**(-alpha)
                r[r2] = a[0] + a[1] * x2 + a[2] * x2**2 + a[3] * x2**3
                r[r3] = x[r3]**(-beta) * t2 ** beta

            return r

        self.lpfactor = lpfactor

        lmin = G.lmax / lpfactor

        if scales is None:
            scales = utils.compute_log_scales(lmin, G.lmax, Nf - 1)
        self.scales = scales

        gb = lambda x: kernel_abspline3(x, 2, 2, 1, 2)
        gl = lambda x: np.exp(-np.power(x, 4))

        lminfac = .4 * lmin

        g = [lambda x: 1.2 * np.exp(-1) * gl(x / lminfac)]
        for i in range(0, Nf - 1):
            g.append(lambda x, i=i: gb(self.scales[i] * x))

        f = lambda x: -gb(x)
        xstar = optimize.minimize_scalar(f, bounds=(1, 2),
                                         method='bounded')
        gamma_l = -f(xstar.x)
        lminfac = .6 * lmin
        g[0] = lambda x: gamma_l * gl(x / lminfac)

        super(Abspline, self).__init__(G, g)

    def _get_extra_repr(self):
        return dict(lpfactor='{:.2f}'.format(self.lpfactor))



# -*- coding: utf-8 -*-

# prevent circular import in Python < 3.5
from pygsp2.graphs import Graph, Grid2d, ImgPatches


class Grid2dImgPatches(Graph):
    r"""Union of a patch graph with a 2D grid graph.

    Parameters
    ----------
    img : array
        Input image.
    aggregate: callable, optional
        Function to aggregate the weights ``Wp`` of the patch graph and the
        ``Wg`` of the grid graph. Default is ``lambda Wp, Wg: Wp + Wg``.
    kwargs : dict
        Parameters passed to :class:`ImgPatches`.

    See Also
    --------
    ImgPatches
    Grid2d

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> from skimage import data, img_as_float
    >>> img = img_as_float(data.camera()[::64, ::64])
    >>> G = graphs.Grid2dImgPatches(img)
    >>> fig, axes = plt.subplots(1, 2)
    >>> _ = axes[0].spy(G.W, markersize=2)
    >>> _ = G.plot(ax=axes[1])

    """

    def __init__(self, img, aggregate=lambda Wp, Wg: Wp + Wg, **kwargs):

        self.Gg = Grid2d(img.shape[0], img.shape[1])
        self.Gp = ImgPatches(img, **kwargs)

        W = aggregate(self.Gp.W, self.Gg.W)
        super(Grid2dImgPatches, self).__init__(W,
                                               coords=self.Gg.coords,
                                               plotting=self.Gg.plotting)

    def _get_extra_repr(self):
        attrs = self.Gg._get_extra_repr()
        attrs.update(self.Gp._get_extra_repr())
        return attrs



# -*- coding: utf-8 -*-

"""
Test suite for the graphs module of the pygsp2 package.

"""

from __future__ import division

import os
import random
import sys
import unittest

import numpy as np
import scipy.linalg
from scipy import sparse
import networkx as nx
import graph_tool as gt
import graph_tool.generation
from skimage import data, img_as_float

from pygsp2 import graphs


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls._G = graphs.Logo()
        cls._G.compute_fourier_basis()
        cls._G.compute_differential_operator()

        cls._rng = np.random.default_rng(42)
        cls._signal = cls._rng.uniform(size=cls._G.N)

        cls._img = img_as_float(data.camera()[::16, ::16])

    @classmethod
    def tearDownClass(cls):
        pass

    def test_graph(self):
        adjacency = [
            [0., 3., 0., 2.],
            [3., 0., 4., 0.],
            [0., 4., 0., 5.],
            [2., 0., 5., 0.],
        ]

        # Input types.
        G = graphs.Graph(adjacency)
        self.assertIs(type(G.W), sparse.csr_matrix)
        adjacency = np.array(adjacency)
        G = graphs.Graph(adjacency)
        self.assertIs(type(G.W), sparse.csr_matrix)
        adjacency = sparse.coo_matrix(adjacency)
        G = graphs.Graph(adjacency)
        self.assertIs(type(G.W), sparse.csr_matrix)
        adjacency = sparse.csr_matrix(adjacency)
        # G = graphs.Graph(adjacency)
        # self.assertIs(G.W, adjacency)  # Not copied if already CSR.

        # Attributes.
        np.testing.assert_allclose(G.W.toarray(), adjacency.toarray())
        np.testing.assert_allclose(G.A.toarray(), G.W.toarray() > 0)
        np.testing.assert_allclose(G.d, np.array([2, 2, 2, 2]))
        np.testing.assert_allclose(G.dw, np.array([5, 7, 9, 7]))
        self.assertEqual(G.n_vertices, 4)
        self.assertIs(G.N, G.n_vertices)
        self.assertEqual(G.n_edges, 4)
        self.assertIs(G.Ne, G.n_edges)

        # Errors and warnings.
        self.assertRaises(ValueError, graphs.Graph, np.ones((3, 4)))
        self.assertRaises(ValueError, graphs.Graph, np.ones((3, 3, 4)))
        self.assertRaises(ValueError, graphs.Graph, [[0, np.nan], [0, 0]])
        self.assertRaises(ValueError, graphs.Graph, [[0, np.inf], [0, 0]])
        with self.assertLogs(level='WARNING'):
            graphs.Graph([[0, -1], [-1, 0]])
        with self.assertLogs(level='WARNING'):
            graphs.Graph([[1, 1], [1, 0]])
        for attr in ['A', 'd', 'dw', 'lmax', 'U', 'e', 'coherence', 'D']:
            # FIXME: The Laplacian L should be there as well.
            self.assertRaises(AttributeError, setattr, G, attr, None)
            self.assertRaises(AttributeError, delattr, G, attr)

    def test_degree(self):
        graph = graphs.Graph([
            [0, 1, 0],
            [1, 0, 2],
            [0, 2, 0],
        ])
        self.assertEqual(graph.is_directed(), False)
        np.testing.assert_allclose(graph.d, [1, 2, 1])
        np.testing.assert_allclose(graph.dw, [1, 3, 2])
        graph = graphs.Graph([
            [0, 1, 0],
            [0, 0, 2],
            [0, 2, 0],
        ])
        self.assertEqual(graph.is_directed(), True)
        np.testing.assert_allclose(graph.d, [0.5, 1.5, 1])
        np.testing.assert_allclose(graph.dw, [0.5, 2.5, 2])

    def test_is_connected(self):
        graph = graphs.Graph([
            [0, 1, 0],
            [1, 0, 2],
            [0, 2, 0],
        ])
        self.assertEqual(graph.is_directed(), False)
        self.assertEqual(graph.is_connected(), True)
        graph = graphs.Graph([
            [0, 1, 0],
            [1, 0, 0],
            [0, 2, 0],
        ])
        self.assertEqual(graph.is_directed(), True)
        self.assertEqual(graph.is_connected(), False)
        graph = graphs.Graph([
            [0, 1, 0],
            [1, 0, 0],
            [0, 0, 0],
        ])
        self.assertEqual(graph.is_directed(), False)
        self.assertEqual(graph.is_connected(), False)
        graph = graphs.Graph([
            [0, 1, 0],
            [0, 0, 2],
            [3, 0, 0],
        ])
        self.assertEqual(graph.is_directed(), True)
        self.assertEqual(graph.is_connected(), True)

    def test_is_directed(self):
        graph = graphs.Graph([
            [0, 3, 0, 0],
            [3, 0, 4, 0],
            [0, 4, 0, 2],
            [0, 0, 2, 0],
        ])
        assert graph.W.nnz == 6
        self.assertEqual(graph.is_directed(), False)
        # In-place modification is not allowed anymore.
        # graph.W[0, 1] = 0
        # assert graph.W.nnz == 6
        # self.assertEqual(graph.is_directed(recompute=True), True)
        # graph.W[1, 0] = 0
        # assert graph.W.nnz == 6
        # self.assertEqual(graph.is_directed(recompute=True), False)

    def test_laplacian(self):

        G = graphs.Graph([
            [0, 3, 0, 1],
            [3, 0, 1, 0],
            [0, 1, 0, 3],
            [1, 0, 3, 0],
        ])
        laplacian = np.array([
            [+4, -3, +0, -1],
            [-3, +4, -1, +0],
            [+0, -1, +4, -3],
            [-1, +0, -3, +4],
        ])
        self.assertFalse(G.is_directed())
        G.compute_laplacian('combinatorial')
        np.testing.assert_allclose(G.L.toarray(), laplacian)
        G.compute_laplacian('normalized')
        np.testing.assert_allclose(G.L.toarray(), laplacian/4)

        G = graphs.Graph([
            [0, 6, 0, 1],
            [0, 0, 0, 0],
            [0, 2, 0, 3],
            [1, 0, 3, 0],
        ])
        self.assertTrue(G.is_directed())
        G.compute_laplacian('combinatorial')
        np.testing.assert_allclose(G.L.toarray(), laplacian)
        G.compute_laplacian('normalized')
        np.testing.assert_allclose(G.L.toarray(), laplacian/4)

        def test_combinatorial(G):
            np.testing.assert_equal(G.L.toarray(), G.L.T.toarray())
            np.testing.assert_equal(G.L.sum(axis=0), 0)
            np.testing.assert_equal(G.L.sum(axis=1), 0)
            np.testing.assert_equal(G.L.diagonal(), G.dw)

        def test_normalized(G):
            np.testing.assert_equal(G.L.toarray(), G.L.T.toarray())
            np.testing.assert_equal(G.L.diagonal(), 1)

        G = graphs.ErdosRenyi(100, directed=False)
        self.assertFalse(G.is_directed())
        G.compute_laplacian(lap_type='combinatorial')
        test_combinatorial(G)
        G.compute_laplacian(lap_type='normalized')
        test_normalized(G)

        G = graphs.ErdosRenyi(100, directed=True)
        self.assertTrue(G.is_directed())
        G.compute_laplacian(lap_type='combinatorial')
        test_combinatorial(G)
        G.compute_laplacian(lap_type='normalized')
        test_normalized(G)

    def test_estimate_lmax(self):

        graph = graphs.Sensor()
        self.assertRaises(ValueError, graph.estimate_lmax, method='unk')

        def check_lmax(graph, lmax):
            graph.estimate_lmax(method='bounds')
            np.testing.assert_allclose(graph.lmax, lmax)
            graph.estimate_lmax(method='lanczos')
            np.testing.assert_allclose(graph.lmax, lmax*1.01)
            graph.compute_fourier_basis()
            np.testing.assert_allclose(graph.lmax, lmax)

        # Full graph (bound is tight).
        n_nodes, value = 10, 2
        adjacency = np.full((n_nodes, n_nodes), value)
        graph = graphs.Graph(adjacency, lap_type='combinatorial')
        check_lmax(graph, lmax=value*n_nodes)

        # Regular bipartite graph (bound is tight).
        adjacency = [
            [0, 0, 1, 1],
            [0, 0, 1, 1],
            [1, 1, 0, 0],
            [1, 1, 0, 0],
        ]
        graph = graphs.Graph(adjacency, lap_type='combinatorial')
        check_lmax(graph, lmax=4)

        # Bipartite graph (bound is tight).
        adjacency = [
            [0, 0, 1, 1],
            [0, 0, 1, 0],
            [1, 1, 0, 0],
            [1, 0, 0, 0],
        ]
        graph = graphs.Graph(adjacency, lap_type='normalized')
        check_lmax(graph, lmax=2)

    def test_fourier_basis(self):
        # Smallest eigenvalue close to zero.
        np.testing.assert_allclose(self._G.e[0], 0, atol=1e-12)
        # First eigenvector is constant.
        N = self._G.N
        np.testing.assert_allclose(self._G.U[:, 0], np.sqrt(N) / N)
        # Control eigenvector direction.
        # assert (self._G.U[0, :] > 0).all()
        # Spectrum bounded by [0, 2] for the normalized Laplacian.
        G = graphs.Logo(lap_type='normalized')
        n = G.N // 2
        # check partial eigendecomposition
        G.compute_fourier_basis(n_eigenvectors=n)
        assert len(G.e) == n
        assert G.U.shape[1] == n
        assert G.e[-1] < 2
        U = G.U
        e = G.e
        # check full eigendecomposition
        G.compute_fourier_basis()
        assert len(G.e) == G.N
        assert G.U.shape[1] == G.N
        assert G.e[-1] < 2
        # eigsh might flip a sign
        np.testing.assert_allclose(np.abs(U), np.abs(G.U[:, :n]),
                                   atol=1e-12)
        np.testing.assert_allclose(e, G.e[:n])

    def test_eigendecompositions(self):
        G = graphs.Logo()
        U1, e1, V1 = scipy.linalg.svd(G.L.toarray())
        U2, e2, V2 = np.linalg.svd(G.L.toarray())
        e3, U3 = np.linalg.eig(G.L.toarray())
        e4, U4 = scipy.linalg.eig(G.L.toarray())
        e5, U5 = np.linalg.eigh(G.L.toarray())
        e6, U6 = scipy.linalg.eigh(G.L.toarray())

        def correct_sign(U):
            signs = np.sign(U[0, :])
            signs[signs == 0] = 1
            return U * signs
        U1 = correct_sign(U1)
        U2 = correct_sign(U2)
        U3 = correct_sign(U3)
        U4 = correct_sign(U4)
        U5 = correct_sign(U5)
        U6 = correct_sign(U6)
        V1 = correct_sign(V1.T)
        V2 = correct_sign(V2.T)

        inds3 = np.argsort(e3)[::-1]
        inds4 = np.argsort(e4)[::-1]
        np.testing.assert_allclose(e2, e1)
        np.testing.assert_allclose(e3[inds3], e1, atol=1e-12)
        np.testing.assert_allclose(e4[inds4], e1, atol=1e-12)
        np.testing.assert_allclose(e5[::-1], e1, atol=1e-12)
        np.testing.assert_allclose(e6[::-1], e1, atol=1e-12)
        np.testing.assert_allclose(U2, U1, atol=1e-12)
        np.testing.assert_allclose(V1, U1, atol=1e-12)
        np.testing.assert_allclose(V2, U1, atol=1e-12)
        np.testing.assert_allclose(U3[:, inds3], U1, atol=1e-10)
        np.testing.assert_allclose(U4[:, inds4], U1, atol=1e-10)
        np.testing.assert_allclose(U5[:, ::-1], U1, atol=1e-10)
        np.testing.assert_allclose(U6[:, ::-1], U1, atol=1e-10)

    def test_fourier_transform(self):
        s = self._rng.uniform(size=(self._G.N, 99, 21))
        s_hat = self._G.gft(s)
        s_star = self._G.igft(s_hat)
        np.testing.assert_allclose(s, s_star, rtol=1e-6)

    def test_edge_list(self):
        for directed in [False, True]:
            G = graphs.ErdosRenyi(100, directed=directed)
            sources, targets, weights = G.get_edge_list()
            if not directed:
                self.assertTrue(np.all(sources <= targets))
            edges = np.arange(G.n_edges)
            np.testing.assert_equal(G.W[sources[edges], targets[edges]],
                                    weights[edges][np.newaxis, :])

    def test_differential_operator(self, n_vertices=98):
        r"""The Laplacian must always be the divergence of the gradient,
        whether the Laplacian is combinatorial or normalized, and whether the
        graph is directed or weighted."""
        def test_incidence_nx(graph):
            r"""Test that the incidence matrix corresponds to NetworkX."""
            incidence_pg = np.sign(graph.D.toarray())
            G = nx.DiGraph if graph.is_directed() else nx.Graph
            graph_nx = nx.from_scipy_sparse_matrix(graph.W, create_using=G)
            incidence_nx = nx.incidence_matrix(graph_nx, oriented=True)
            np.testing.assert_equal(incidence_pg, incidence_nx.toarray())
        for graph in [graphs.Graph(np.zeros((n_vertices, n_vertices))),
                      graphs.Graph(np.identity(n_vertices)),
                      graphs.Graph([[0, 0.8], [0.8, 0]]),
                      graphs.Graph([[1.3, 0], [0.4, 0.5]]),
                      graphs.ErdosRenyi(n_vertices, directed=False, seed=42),
                      graphs.ErdosRenyi(n_vertices, directed=True, seed=42)]:
            for lap_type in ['combinatorial', 'normalized']:
                graph.compute_laplacian(lap_type)
                graph.compute_differential_operator()
                L = graph.D.dot(graph.D.T)
                np.testing.assert_allclose(L.toarray(), graph.L.toarray())
                test_incidence_nx(graph)

    def test_difference(self):
        for lap_type in ['combinatorial', 'normalized']:
            G = graphs.Logo(lap_type=lap_type)
            y = G.grad(self._signal)
            self.assertEqual(len(y), G.n_edges)
            z = G.div(y)
            self.assertEqual(len(z), G.n_vertices)
            np.testing.assert_allclose(z, G.L.dot(self._signal))

    def test_dirichlet_energy(self, n_vertices=100):
        r"""The Dirichlet energy is defined as the norm of the gradient."""
        signal = np.random.default_rng(42).uniform(size=n_vertices)
        for lap_type in ['combinatorial', 'normalized']:
            graph = graphs.BarabasiAlbert(n_vertices, lap_type=lap_type)
            graph.compute_differential_operator()
            energy = graph.dirichlet_energy(signal)
            grad_norm = np.sum(graph.grad(signal)**2)
            np.testing.assert_allclose(energy, grad_norm)

    def test_empty_graph(self, n_vertices=11):
        """Empty graphs have either no edge, or self-loops only. The Laplacian
        doesn't see self-loops, as the gradient on those edges is always zero.
        """
        adjacencies = [
            np.zeros((n_vertices, n_vertices)),
            np.identity(n_vertices),
        ]
        for adjacency, n_edges in zip(adjacencies, [0, n_vertices]):
            graph = graphs.Graph(adjacency)
            self.assertEqual(graph.n_vertices, n_vertices)
            self.assertEqual(graph.n_edges, n_edges)
            self.assertEqual(graph.W.nnz, n_edges)
            for laplacian in ['combinatorial', 'normalized']:
                graph.compute_laplacian(laplacian)
                self.assertEqual(graph.L.nnz, 0)
                sources, targets, weights = graph.get_edge_list()
                self.assertEqual(len(sources), n_edges)
                self.assertEqual(len(targets), n_edges)
                self.assertEqual(len(weights), n_edges)
                graph.compute_differential_operator()
                self.assertEqual(graph.D.nnz, 0)
                graph.compute_fourier_basis()
                np.testing.assert_allclose(graph.U, np.identity(n_vertices))
                np.testing.assert_allclose(graph.e, np.zeros(n_vertices))
            # NetworkX uses the same conventions.
            G = nx.from_scipy_sparse_matrix(graph.W)
            self.assertEqual(nx.laplacian_matrix(G).nnz, 0)
            self.assertEqual(nx.normalized_laplacian_matrix(G).nnz, 0)
            self.assertEqual(nx.incidence_matrix(G).nnz, 0)

    def test_adjacency_types(self, n_vertices=10):

        rng = np.random.default_rng(42)
        W = 10 * np.abs(rng.normal(size=(n_vertices, n_vertices)))
        W = W + W.T
        W = W - np.diag(np.diag(W))

        def test(adjacency):
            G = graphs.Graph(adjacency)
            G.compute_laplacian('combinatorial')
            G.compute_laplacian('normalized')
            G.estimate_lmax()
            G.compute_fourier_basis()
            G.compute_differential_operator()

        test(W)
        test(W.astype(np.float32))
        test(W.astype(int))
        test(sparse.csr_matrix(W))
        test(sparse.csr_matrix(W, dtype=np.float32))
        test(sparse.csr_matrix(W, dtype=int))
        test(sparse.csc_matrix(W))
        test(sparse.coo_matrix(W))

    def test_set_signal(self, name='test'):
        signal = np.zeros(self._G.n_vertices)
        self._G.set_signal(signal, name)
        self.assertIs(self._G.signals[name], signal)
        signal = np.zeros(self._G.n_vertices // 2)
        self.assertRaises(ValueError, self._G.set_signal, signal, name)

    def test_set_coordinates(self):
        G = graphs.FullConnected()
        coords = self._rng.uniform(size=(G.N, 2))
        G.set_coordinates(coords)
        G.set_coordinates('ring2D')
        G.set_coordinates('random2D')
        G.set_coordinates('random3D')
        G.set_coordinates('spring')
        G.set_coordinates('spring', dim=3)
        G.set_coordinates('spring', dim=3, pos=G.coords)
        G.set_coordinates('laplacian_eigenmap2D')
        G.set_coordinates('laplacian_eigenmap3D')
        self.assertRaises(AttributeError, G.set_coordinates, 'community2D')
        G = graphs.Community()
        G.set_coordinates('community2D')
        self.assertRaises(ValueError, G.set_coordinates, 'invalid')

    def test_line_graph(self):
        adjacency = [
            [0, 1, 1, 3],
            [1, 0, 1, 0],
            [1, 1, 0, 1],
            [3, 0, 1, 0],
        ]
        coords = [
            [0, 0],
            [4, 0],
            [4, 2],
            [0, 2],
        ]
        graph = graphs.Graph(adjacency, coords=coords)
        graph = graphs.LineGraph(graph)
        adjacency = [
            [0, 1, 1, 1, 0],
            [1, 0, 1, 1, 1],
            [1, 1, 0, 0, 1],
            [1, 1, 0, 0, 1],
            [0, 1, 1, 1, 0],
        ]
        coords = [
            [2, 0],
            [2, 1],
            [0, 1],
            [4, 1],
            [2, 2],
        ]
        np.testing.assert_equal(graph.W.toarray(), adjacency)
        np.testing.assert_equal(graph.coords, coords)
        with self.assertLogs(level='WARNING'):
            graphs.LineGraph(graphs.Graph([[0, 2], [2, 0]]))

    def test_subgraph(self, n_vertices=100):
        self._G.set_signal(self._G.coords, 'coords')
        graph = self._G.subgraph(range(n_vertices))
        self.assertEqual(graph.n_vertices, n_vertices)
        self.assertEqual(graph.coords.shape, (n_vertices, 2))
        self.assertEqual(graph.signals['coords'].shape, (n_vertices, 2))
        self.assertIs(graph.lap_type, self._G.lap_type)
        self.assertEqual(graph.plotting, self._G.plotting)

    def test_nngraph(self, n_vertices=30):
        rng = np.random.default_rng(42)
        Xin = rng.normal(size=(n_vertices, 3))
        dist_types = ['euclidean', 'manhattan', 'max_dist', 'minkowski']

        for dist_type in dist_types:

            # Only p-norms with 1<=p<=infinity permitted.
            if dist_type != 'minkowski':
                graphs.NNGraph(Xin, NNtype='radius', dist_type=dist_type)
                graphs.NNGraph(Xin, NNtype='knn', dist_type=dist_type)

            # Distance type unsupported in the C bindings,
            # use the C++ bindings instead.
            if dist_type != 'max_dist':
                graphs.NNGraph(Xin, use_flann=True, NNtype='knn',
                               dist_type=dist_type)

    def test_bunny(self):
        graphs.Bunny()

    def test_cube(self):
        graphs.Cube()
        graphs.Cube(nb_dim=2)

    def test_sphere(self):
        graphs.Sphere()

    def test_twomoons(self):
        graphs.TwoMoons(moontype='standard')
        graphs.TwoMoons(moontype='synthesized')

    def test_torus(self):
        graphs.Torus()

    def test_comet(self, n=100, k=10):
        graph = graphs.Comet(n, k)
        self.assertEqual(graph.n_vertices, n)
        self.assertEqual(graph.n_edges, n-1)
        self.assertEqual(graph.dw[0], k)
        graph = graphs.Comet(7, 4)
        adjacency = [
            [0, 1, 1, 1, 1, 0, 0],  # center
            [1, 0, 0, 0, 0, 0, 0],  # branch
            [1, 0, 0, 0, 0, 0, 0],  # branch
            [1, 0, 0, 0, 0, 0, 0],  # branch
            [1, 0, 0, 0, 0, 1, 0],  # tail
            [0, 0, 0, 0, 1, 0, 1],  # tail
            [0, 0, 0, 0, 0, 1, 0],  # tail
        ]
        np.testing.assert_array_equal(graph.W.toarray(), adjacency)
        coords = [[0, 0], [0, 1], [-1, 0], [0, -1], [1, 0], [2, 0], [3, 0]]
        np.testing.assert_allclose(graph.coords, coords, atol=1e-10)
        # Comet generalizes Path.
        g1 = graphs.Comet(n, 0)
        g2 = graphs.Path(n)
        np.testing.assert_array_equal(g1.W.toarray(), g2.W.toarray())
        # Comet generalizes Star.
        g1 = graphs.Comet(n, n-1)
        g2 = graphs.Star(n)
        np.testing.assert_array_equal(g1.W.toarray(), g2.W.toarray())

    def test_star(self, n=20):
        graph = graphs.Star(n)
        self.assertEqual(graph.n_vertices, n)
        self.assertEqual(graph.n_edges, n-1)
        np.testing.assert_array_equal(graph.d, [n-1] + (n-1) * [1])
        np.testing.assert_allclose(np.linalg.norm(graph.coords[1:], axis=1), 1)

    def test_lowstretchtree(self):
        graphs.LowStretchTree()

    def test_randomregular(self):
        k = 6
        G = graphs.RandomRegular(k=k)
        np.testing.assert_equal(G.W.sum(0), k)
        np.testing.assert_equal(G.W.sum(1), k)

    def test_ring(self):
        graphs.Ring()
        graphs.Ring(N=32, k=16)
        self.assertRaises(ValueError, graphs.Ring, 2)
        self.assertRaises(ValueError, graphs.Ring, 5, k=3)

    def test_community(self):
        graphs.Community()
        graphs.Community(comm_density=0.2)
        graphs.Community(k_neigh=5)
        graphs.Community(N=100, Nc=3, comm_sizes=[20, 50, 30])

    def test_minnesota(self):
        graphs.Minnesota()

    def test_sensor(self):
        graphs.Sensor(3000)
        graphs.Sensor(N=100, distributed=True)
        self.assertRaises(ValueError, graphs.Sensor, N=101, distributed=True)
        graphs.Sensor(N=101, distributed=False)
        graphs.Sensor(seed=10)
        graphs.Sensor(k=20)

    def test_stochasticblockmodel(self):
        graphs.StochasticBlockModel(N=100, directed=True)
        graphs.StochasticBlockModel(N=100, directed=False)
        graphs.StochasticBlockModel(N=100, self_loops=True)
        graphs.StochasticBlockModel(N=100, self_loops=False)
        graphs.StochasticBlockModel(N=100, connected=True, seed=42)
        graphs.StochasticBlockModel(N=100, connected=False)
        self.assertRaises(ValueError, graphs.StochasticBlockModel,
                          N=100, p=0, q=0, connected=True)

    def test_airfoil(self):
        graphs.Airfoil()

    def test_davidsensornet(self):
        graphs.DavidSensorNet()
        graphs.DavidSensorNet(N=500)
        graphs.DavidSensorNet(N=128)

    def test_erdosreny(self):
        graphs.ErdosRenyi(N=100, connected=False, directed=False)
        graphs.ErdosRenyi(N=100, connected=False, directed=True)
        graphs.ErdosRenyi(N=100, connected=True, directed=False, seed=42)
        graphs.ErdosRenyi(N=100, connected=True, directed=True, seed=42)
        G = graphs.ErdosRenyi(N=100, p=1, self_loops=True)
        self.assertEqual(G.W.nnz, 100**2)

    def test_fullconnected(self):
        graphs.FullConnected()

    def test_logo(self):
        graphs.Logo()

    def test_path(self, n=5):
        graph = graphs.Path(n, directed=False)
        adjacency = [
            [0, 1, 0, 0, 0],
            [1, 0, 1, 0, 0],
            [0, 1, 0, 1, 0],
            [0, 0, 1, 0, 1],
            [0, 0, 0, 1, 0],
        ]
        coords = [[0, 0], [1, 0], [2, 0], [3, 0], [4, 0]]
        np.testing.assert_array_equal(graph.W.toarray(), adjacency)
        np.testing.assert_array_equal(graph.coords, coords)
        graph = graphs.Path(n, directed=True)
        adjacency = [
            [0, 1, 0, 0, 0],
            [0, 0, 1, 0, 0],
            [0, 0, 0, 1, 0],
            [0, 0, 0, 0, 1],
            [0, 0, 0, 0, 0],
        ]
        np.testing.assert_array_equal(graph.W.toarray(), adjacency)
        np.testing.assert_array_equal(graph.coords, coords)

    def test_randomring(self):
        graphs.RandomRing()
        G = graphs.RandomRing(angles=[0, 2, 1])
        self.assertEqual(G.N, 3)
        self.assertRaises(ValueError, graphs.RandomRing, 2)
        self.assertRaises(ValueError, graphs.RandomRing, angles=[0, 2])
        self.assertRaises(ValueError, graphs.RandomRing, angles=[0, 2, 7])
        self.assertRaises(ValueError, graphs.RandomRing, angles=[0, 2, -1])

    def test_swissroll(self):
        graphs.SwissRoll(srtype='uniform')
        graphs.SwissRoll(srtype='classic')
        graphs.SwissRoll(noise=True)
        graphs.SwissRoll(noise=False)
        graphs.SwissRoll(dim=2)
        graphs.SwissRoll(dim=3)

    def test_grid2d(self):
        graphs.Grid2d(3, 2)
        graphs.Grid2d(3)

    def test_imgpatches(self):
        graphs.ImgPatches(img=self._img, patch_shape=(3, 3))

    def test_grid2dimgpatches(self):
        graphs.Grid2dImgPatches(img=self._img, patch_shape=(3, 3))

    def test_grid2d_diagonals(self):
        value = 0.5
        G = graphs.Grid2d(6, 7, diagonal=value)
        self.assertEqual(G.W[2, 8], value)
        self.assertEqual(G.W[9, 1], value)
        self.assertEqual(G.W[9, 3], value)
        self.assertEqual(G.W[2, 14], 0.0)
        self.assertEqual(G.W[17, 1], 0.0)
        self.assertEqual(G.W[9, 16], 1.0)
        self.assertEqual(G.W[20, 27], 1.0)

suite_graphs = unittest.TestLoader().loadTestsFromTestCase(TestCase)


class TestImportExport(unittest.TestCase):

    def test_networkx_export_import(self):
        # Export to networkx and reimport to PyGSP2

        # Exporting the Bunny graph
        g = graphs.Bunny()
        g_nx = g.to_networkx()
        g2 = graphs.Graph.from_networkx(g_nx)
        np.testing.assert_array_equal(g.W.todense(), g2.W.todense())

    def test_networkx_import_export(self):
        # Import from networkx then export to networkx again
        g_nx = nx.gnm_random_graph(100, 50)  # Generate a random graph
        g = graphs.Graph.from_networkx(g_nx).to_networkx()

        np.testing.assert_array_equal(nx.adjacency_matrix(g_nx).todense(),
                                      nx.adjacency_matrix(g).todense())

    def test_graphtool_export_import(self):
        # Export to graph tool and reimport to PyGSP2 directly
        # The exported graph is a simple one without an associated Signal
        g = graphs.Bunny()
        g_gt = g.to_graphtool()
        g2 = graphs.Graph.from_graphtool(g_gt)
        np.testing.assert_array_equal(g.W.todense(), g2.W.todense())

    def test_graphtool_multiedge_import(self):
        # Manualy create a graph with multiple edges
        g_gt = gt.Graph()
        g_gt.add_vertex(n=10)
        # connect edge (3,6) three times
        for i in range(3):
            g_gt.add_edge(g_gt.vertex(3), g_gt.vertex(6))
        g = graphs.Graph.from_graphtool(g_gt)
        self.assertEqual(g.W[3, 6], 3.0)

        eprop_double = g_gt.new_edge_property("double")

        # Set the weight of 2 out of the 3 edges. The last one has a default weight of 0
        e = g_gt.edge(3, 6, all_edges=True)
        eprop_double[e[0]] = 8.0
        eprop_double[e[1]] = 1.0

        g_gt.edge_properties["weight"] = eprop_double
        g3 = graphs.Graph.from_graphtool(g_gt)
        self.assertEqual(g3.W[3, 6], 9.0)

    def test_graphtool_import_export(self):
        # Import to PyGSP2 and export again to graph tool directly
        # create a random graphTool graph that does not contain multiple edges and no signal
        rng = np.random.default_rng(42)
        graph_gt = gt.generation.random_graph(100, lambda: (rng.poisson(4), rng.poisson(4)))

        eprop_double = graph_gt.new_edge_property("double")
        for e in graph_gt.edges():
            eprop_double[e] = random.random()
        graph_gt.edge_properties["weight"] = eprop_double

        graph2_gt = graphs.Graph.from_graphtool(graph_gt).to_graphtool()

        self.assertEqual(graph_gt.num_edges(), graph2_gt.num_edges(),
                         "the number of edges does not correspond")

        def key(edge): return str(edge.source()) + ":" + str(edge.target())

        for e1, e2 in zip(sorted(graph_gt.edges(), key=key), sorted(graph2_gt.edges(), key=key)):
            self.assertEqual(e1.source(), e2.source())
            self.assertEqual(e1.target(), e2.target())
        for v1, v2 in zip(graph_gt.vertices(), graph2_gt.vertices()):
            self.assertEqual(v1, v2)

    def test_networkx_signal_export(self):
        graph = graphs.BarabasiAlbert(N=100, seed=42)
        rng = np.random.default_rng(42)
        signal1 = rng.normal(0, 1, graph.N)
        signal2 = rng.integers(0, 10, graph.N)
        graph.set_signal(signal1, "signal1")
        graph.set_signal(signal2, "signal2")
        graph_nx = graph.to_networkx()
        for i in range(graph.N):
            self.assertEqual(graph_nx.nodes[i]["signal1"], signal1[i])
            self.assertEqual(graph_nx.nodes[i]["signal2"], signal2[i])
        # invalid signal type
        graph = graphs.Path(3)
        graph.set_signal(np.array(['a', 'b', 'c']), 'sig')
        self.assertRaises(ValueError, graph.to_networkx)

    def test_graphtool_signal_export(self):
        g = graphs.Logo()
        rng = np.random.default_rng(42)
        s = rng.normal(0, 1, size=g.N)
        s2 = rng.integers(0, 10, size=g.N)
        g.set_signal(s, "signal1")
        g.set_signal(s2, "signal2")
        g_gt = g.to_graphtool()
        # Check the signals on all nodes
        for i, v in enumerate(g_gt.vertices()):
            self.assertEqual(g_gt.vertex_properties["signal1"][v], s[i])
            self.assertEqual(g_gt.vertex_properties["signal2"][v], s2[i])
        # invalid signal type
        graph = graphs.Path(3)
        graph.set_signal(np.array(['a', 'b', 'c']), 'sig')
        self.assertRaises(TypeError, graph.to_graphtool)

    def test_graphtool_signal_import(self):
        g_gt = gt.Graph()
        g_gt.add_vertex(10)

        g_gt.add_edge(g_gt.vertex(3), g_gt.vertex(6))
        g_gt.add_edge(g_gt.vertex(4), g_gt.vertex(6))
        g_gt.add_edge(g_gt.vertex(7), g_gt.vertex(2))

        vprop_double = g_gt.new_vertex_property("double")

        vprop_double[g_gt.vertex(0)] = 5
        vprop_double[g_gt.vertex(1)] = -3
        vprop_double[g_gt.vertex(2)] = 2.4

        g_gt.vertex_properties["signal"] = vprop_double
        g = graphs.Graph.from_graphtool(g_gt)
        self.assertEqual(g.signals["signal"][0], 5.0)
        self.assertEqual(g.signals["signal"][1], -3.0)
        self.assertEqual(g.signals["signal"][2], 2.4)

    def test_networkx_signal_import(self):
        graph_nx = nx.Graph()
        graph_nx.add_nodes_from(range(2, 5))
        graph_nx.add_edges_from([(3, 4), (2, 4), (3, 5)])
        nx.set_node_attributes(graph_nx, {2: 4, 3: 5, 5: 2.3}, "s")
        graph_pg = graphs.Graph.from_networkx(graph_nx)
        np.testing.assert_allclose(graph_pg.signals["s"], [4, 5, np.nan, 2.3])

    def test_no_weights(self):

        adjacency = np.array([[0, 1, 0], [1, 0, 1], [0, 1, 0]])

        # NetworkX no weights.
        graph_nx = nx.Graph()
        graph_nx.add_edge(0, 1)
        graph_nx.add_edge(1, 2)
        graph_pg = graphs.Graph.from_networkx(graph_nx)
        np.testing.assert_allclose(graph_pg.W.toarray(), adjacency)

        # NetworkX non-existent weight name.
        graph_nx.edges[(0, 1)]['weight'] = 2
        graph_nx.edges[(1, 2)]['weight'] = 2
        graph_pg = graphs.Graph.from_networkx(graph_nx)
        np.testing.assert_allclose(graph_pg.W.toarray(), 2*adjacency)
        graph_pg = graphs.Graph.from_networkx(graph_nx, weight='unknown')
        np.testing.assert_allclose(graph_pg.W.toarray(), adjacency)

        # Graph-tool no weights.
        graph_gt = gt.Graph(directed=False)
        graph_gt.add_edge(0, 1)
        graph_gt.add_edge(1, 2)
        graph_pg = graphs.Graph.from_graphtool(graph_gt)
        np.testing.assert_allclose(graph_pg.W.toarray(), adjacency)

        # Graph-tool non-existent weight name.
        prop = graph_gt.new_edge_property("double")
        prop[graph_gt.edge(0, 1)] = 2
        prop[graph_gt.edge(1, 2)] = 2
        graph_gt.edge_properties["weight"] = prop
        graph_pg = graphs.Graph.from_graphtool(graph_gt)
        np.testing.assert_allclose(graph_pg.W.toarray(), 2*adjacency)
        graph_pg = graphs.Graph.from_graphtool(graph_gt, weight='unknown')
        np.testing.assert_allclose(graph_pg.W.toarray(), adjacency)

    def test_break_join_signals(self):
        """Multi-dim signals are broken on export and joined on import."""
        graph1 = graphs.Sensor(20, seed=42)
        graph1.set_signal(graph1.coords, 'coords')
        # networkx
        graph2 = graph1.to_networkx()
        graph2 = graphs.Graph.from_networkx(graph2)
        np.testing.assert_allclose(graph2.signals['coords'], graph1.coords)
        # graph-tool
        graph2 = graph1.to_graphtool()
        graph2 = graphs.Graph.from_graphtool(graph2)
        np.testing.assert_allclose(graph2.signals['coords'], graph1.coords)
        # save and load (need ordered dicts)
        filename = 'graph.graphml'
        graph1.save(filename)
        graph2 = graphs.Graph.load(filename)
        np.testing.assert_allclose(graph2.signals['coords'], graph1.coords)
        os.remove(filename)

    def test_save_load(self):

        # TODO: test with multiple graphs and signals
        # * dtypes (float, int, bool) of adjacency and signals
        # * empty graph / isolated nodes

        G1 = graphs.Sensor(seed=42)
        W = G1.W.toarray()
        sig = np.random.default_rng(42).normal(size=G1.N)
        G1.set_signal(sig, 's')

        for fmt in ['graphml', 'gml', 'gexf']:
            for backend in ['networkx', 'graph-tool']:

                if fmt == 'gexf' and backend == 'graph-tool':
                    self.assertRaises(ValueError, G1.save, 'g', fmt, backend)
                    self.assertRaises(ValueError, graphs.Graph.load, 'g', fmt,
                                      backend)
                    os.remove('g')
                    continue

                atol = 1e-5 if fmt == 'gml' and backend == 'graph-tool' else 0

                for filename, fmt in [('graph.' + fmt, None), ('graph', fmt)]:
                    G1.save(filename, fmt, backend)
                    G2 = graphs.Graph.load(filename, fmt, backend)
                    np.testing.assert_allclose(G2.W.toarray(), W, atol=atol)
                    np.testing.assert_allclose(G2.signals['s'], sig, atol=atol)
                    os.remove(filename)

        self.assertRaises(ValueError, graphs.Graph.load, 'g.gml', fmt='?')
        self.assertRaises(ValueError, graphs.Graph.load, 'g.gml', backend='?')
        self.assertRaises(ValueError, G1.save, 'g.gml', fmt='?')
        self.assertRaises(ValueError, G1.save, 'g.gml', backend='?')

    def test_import_errors(self):
        from unittest.mock import patch
        graph = graphs.Sensor()
        filename = 'graph.gml'
        with patch.dict(sys.modules, {'networkx': None}):
            self.assertRaises(ImportError, graph.to_networkx)
            self.assertRaises(ImportError, graphs.Graph.from_networkx, None)
            self.assertRaises(ImportError, graph.save, filename,
                              backend='networkx')
            self.assertRaises(ImportError, graphs.Graph.load, filename,
                              backend='networkx')
            graph.save(filename)
            graphs.Graph.load(filename)
        with patch.dict(sys.modules, {'graph_tool': None}):
            self.assertRaises(ImportError, graph.to_graphtool)
            self.assertRaises(ImportError, graphs.Graph.from_graphtool, None)
            self.assertRaises(ImportError, graph.save, filename,
                              backend='graph-tool')
            self.assertRaises(ImportError, graphs.Graph.load, filename,
                              backend='graph-tool')
            graph.save(filename)
            graphs.Graph.load(filename)
        with patch.dict(sys.modules, {'networkx': None, 'graph_tool': None}):
            self.assertRaises(ImportError, graph.save, filename)
            self.assertRaises(ImportError, graphs.Graph.load, filename)
        os.remove(filename)


suite_import_export = unittest.TestLoader().loadTestsFromTestCase(TestImportExport)
suite = unittest.TestSuite([suite_graphs, suite_import_export])



# -*- coding: utf-8 -*-

import numpy as np


class LayoutMixIn(object):

    def set_coordinates(self, kind='spring', seed=None, **kwargs):
        r"""Set node's coordinates (their position when plotting).

        Parameters
        ----------
        kind : string or array_like
            Kind of coordinates to generate. It controls the position of the
            nodes when plotting the graph. Can either pass an array of size Nx2
            or Nx3 to set the coordinates manually or the name of a layout
            algorithm. Available algorithms: community2D, random2D, random3D,
            ring2D, line1D, spring, laplacian_eigenmap2D, laplacian_eigenmap3D.
            Default is 'spring'.
        seed : int
            Seed for the random number generator when `kind` is `'random'`,
            `'community'`, or `'spring'`.
        kwargs : dict
            Additional parameters to be passed to the Fruchterman-Reingold
            force-directed algorithm when kind is spring.

        Examples
        --------
        >>> G = graphs.ErdosRenyi()
        >>> G.set_coordinates()
        >>> fig, ax = G.plot()

        """

        if not isinstance(kind, str):
            coords = np.asanyarray(kind).squeeze()
            check_1d = (coords.ndim == 1)
            check_2d_3d = (coords.ndim == 2) and (2 <= coords.shape[1] <= 3)
            if coords.shape[0] != self.N or not (check_1d or check_2d_3d):
                raise ValueError('Expecting coordinates to be of size N, Nx2, '
                                 'or Nx3.')
            self.coords = coords

        elif kind == 'line1D':
            self.coords = np.arange(self.N)

        elif kind == 'line2D':
            x, y = np.arange(self.N), np.zeros(self.N)
            self.coords = np.stack([x, y], axis=1)

        elif kind == 'ring2D':
            angle = np.arange(self.N) * 2 * np.pi / self.N
            self.coords = np.stack([np.cos(angle), np.sin(angle)], axis=1)

        elif kind == 'random2D':
            rng = np.random.default_rng(seed)
            self.coords = rng.uniform(size=(self.N, 2))

        elif kind == 'random3D':
            rng = np.random.default_rng(seed)
            self.coords = rng.uniform(size=(self.N, 3))

        elif kind == 'spring':
            self.coords = self._fruchterman_reingold(seed=seed, **kwargs)

        elif kind == 'community2D':
            if not hasattr(self, 'info') or 'node_com' not in self.info:
                ValueError('Missing arguments to the graph to be able to '
                           'compute community coordinates.')

            if 'world_rad' not in self.info:
                self.info['world_rad'] = np.sqrt(self.N)

            if 'comm_sizes' not in self.info:
                counts = Counter(self.info['node_com'])
                self.info['comm_sizes'] = np.array([cnt[1] for cnt
                                                    in sorted(counts.items())])

            Nc = self.info['comm_sizes'].shape[0]

            self.info['com_coords'] = self.info['world_rad'] * \
                np.array(list(zip(
                    np.cos(2 * np.pi * np.arange(1, Nc + 1) / Nc),
                    np.sin(2 * np.pi * np.arange(1, Nc + 1) / Nc))))

            # Coordinates of the nodes inside their communities
            rng = np.random.default_rng(seed)
            coords = rng.uniform(size=(self.N, 2))
            self.coords = np.array([[elem[0] * np.cos(2 * np.pi * elem[1]),
                                     elem[0] * np.sin(2 * np.pi * elem[1])]
                                    for elem in coords])

            for i in range(self.N):
                # Set coordinates as an offset from the center of the community
                # it belongs to
                comm_idx = self.info['node_com'][i]
                comm_rad = np.sqrt(self.info['comm_sizes'][comm_idx])
                self.coords[i] = self.info['com_coords'][comm_idx] + \
                    comm_rad * self.coords[i]
        elif kind == 'laplacian_eigenmap2D':
            self.compute_fourier_basis(n_eigenvectors=3)
            self.coords = self.U[:, 1:3]
        elif kind == 'laplacian_eigenmap3D':
            self.compute_fourier_basis(n_eigenvectors=4)
            self.coords = self.U[:, 1:4]
        else:
            raise ValueError('Unexpected argument kind={}.'.format(kind))

    def _fruchterman_reingold(self, dim=2, k=None, pos=None, fixed=[],
                              iterations=50, scale=1.0, center=None,
                              seed=None):
        # TODO doc
        # fixed: list of nodes with fixed coordinates
        # Position nodes using Fruchterman-Reingold force-directed algorithm.

        if center is None:
            center = np.zeros((1, dim))

        if np.shape(center)[1] != dim:
            self.logger.error('Spring coordinates: center has wrong size.')
            center = np.zeros((1, dim))

        if pos is None:
            dom_size = 1
            pos_arr = None
        else:
            # Determine size of existing domain to adjust initial positions
            dom_size = np.max(pos)
            rng = np.random.default_rng(seed)
            pos_arr = rng.uniform(size=(self.N, dim))
            pos_arr = pos_arr * dom_size + center
            for i in range(self.N):
                pos_arr[i] = np.asanyarray(pos[i])

        if k is None and len(fixed) > 0:
            # We must adjust k by domain size for layouts that are not near 1x1
            k = dom_size / np.sqrt(self.N)

        pos = _sparse_fruchterman_reingold(self.A, dim, k, pos_arr,
                                           fixed, iterations, seed)

        if len(fixed) == 0:
            pos = _rescale_layout(pos, scale=scale) + center

        return pos


def _sparse_fruchterman_reingold(A, dim, k, pos, fixed, iterations, seed):
    # Position nodes in adjacency matrix A using Fruchterman-Reingold
    nnodes = A.shape[0]

    # make sure we have a LIst of Lists representation
    try:
        A = A.tolil()
    except Exception:
        A = (sparse.coo_matrix(A)).tolil()

    if pos is None:
        # random initial positions
        rng = np.random.default_rng(seed)
        pos = rng.uniform(size=(nnodes, dim))

    # optimal distance between nodes
    if k is None:
        k = np.sqrt(1.0 / nnodes)

    # simple cooling scheme.
    # linearly step down by dt on each iteration so last iteration is size dt.
    t = 0.1
    dt = t / float(iterations + 1)

    displacement = np.zeros((dim, nnodes))
    for iteration in range(iterations):
        displacement *= 0
        # loop over rows
        for i in range(nnodes):
            if i in fixed:
                continue
            # difference between this row's node position and all others
            delta = (pos[i] - pos).T
            # distance between points
            distance = np.sqrt((delta**2).sum(axis=0))
            # enforce minimum distance of 0.01
            distance = np.where(distance < 0.01, 0.01, distance)
            # the adjacency matrix row
            Ai = A[i, :].toarray()
            # displacement "force"
            displacement[:, i] += \
                (delta * (k * k / distance**2 - Ai * distance / k)).sum(axis=1)
        # update positions
        length = np.sqrt((displacement**2).sum(axis=0))
        length = np.where(length < 0.01, 0.1, length)
        pos += (displacement * t / length).T
        # cool temperature
        t -= dt

    return pos


def _rescale_layout(pos, scale=1):
    # rescale to (-scale, scale) in all axes

    # shift origin to (0,0)
    lim = 0  # max coordinate for all axes
    for i in range(pos.shape[1]):
        pos[:, i] -= pos[:, i].mean()
        lim = max(pos[:, i].max(), lim)
    # rescale to (-scale,scale) in all directions, preserves aspect
    for i in range(pos.shape[1]):
        pos[:, i] *= scale / lim
    return pos



# -*- coding: utf-8 -*-

import numpy as np

from pygsp2 import utils
from . import Filter  # prevent circular import in Python < 3.5


class Meyer(Filter):
    r"""Design a filter bank of Meyer wavelets (tight frame).

    Parameters
    ----------
    G : graph
    Nf : int
        Number of filters from 0 to lmax (default = 6).
    scales : ndarray
        Vector of scales to be used (default: log scale).

    References
    ----------
    Use of this kernel for SGWT proposed by Nora Leonardi and Dimitri Van De
    Ville in :cite:`leonardi2011wavelet`.

    Examples
    --------

    Filter bank's representation in Fourier and time (ring graph) domains.

    >>> import matplotlib.pyplot as plt
    >>> G = graphs.Ring(N=20)
    >>> G.estimate_lmax()
    >>> G.set_coordinates('line1D')
    >>> g = filters.Meyer(G)
    >>> s = g.localize(G.N // 2)
    >>> fig, axes = plt.subplots(1, 2)
    >>> _ = g.plot(ax=axes[0])
    >>> _ = G.plot(s, ax=axes[1])

    """

    def __init__(self, G, Nf=6, scales=None):

        if scales is None:
            scales = (4./(3 * G.lmax)) * np.power(2., np.arange(Nf-2, -1, -1))
        self.scales = scales

        if len(scales) != Nf - 1:
            raise ValueError('len(scales) should be Nf-1.')

        kernels = [lambda x: kernel(scales[0] * x, 'scaling_function')]

        for i in range(Nf - 1):
            kernels.append(lambda x, i=i: kernel(scales[i] * x, 'wavelet'))

        def kernel(x, kernel_type):
            r"""
            Evaluates Meyer function and scaling function

            * meyer wavelet kernel: supported on [2/3,8/3]
            * meyer scaling function kernel: supported on [0,4/3]
            """

            x = np.asanyarray(x)

            l1 = 2/3.
            l2 = 4/3.  # 2*l1
            l3 = 8/3.  # 4*l1

            def v(x):
                return x**4 * (35 - 84*x + 70*x**2 - 20*x**3)

            r1ind = (x < l1)
            r2ind = (x >= l1) * (x < l2)
            r3ind = (x >= l2) * (x < l3)

            # as we initialize r with zero, computed function will implicitly
            # be zero for all x not in one of the three regions defined above
            r = np.zeros(x.shape)
            if kernel_type == 'scaling_function':
                r[r1ind] = 1
                r[r2ind] = np.cos((np.pi/2) * v(np.abs(x[r2ind])/l1 - 1))
            elif kernel_type == 'wavelet':
                r[r2ind] = np.sin((np.pi/2) * v(np.abs(x[r2ind])/l1 - 1))
                r[r3ind] = np.cos((np.pi/2) * v(np.abs(x[r3ind])/l2 - 1))
            else:
                raise ValueError('Unknown kernel type {}'.format(kernel_type))

            return r

        super(Meyer, self).__init__(G, kernels)



# -*- coding: utf-8 -*-

import numpy as np
from scipy import sparse

from . import Graph  # prevent circular import in Python < 3.5


class BarabasiAlbert(Graph):
    r"""Barabasi-Albert preferential attachment.

    The Barabasi-Albert graph is constructed by connecting nodes in two steps.
    First, m0 nodes are created. Then, nodes are added one by one.

    By lack of clarity, we take the liberty to create it as follows:

        1. the m0 initial nodes are disconnected,
        2. each node is connected to m of the older nodes with a probability
           distribution depending of the node-degrees of the other nodes,
           :math:`p_n(i) = \frac{1 + k_i}{\sum_j{1 + k_j}}`.

    Parameters
    ----------
    N : int
        Number of nodes (default is 1000)
    m0 : int
        Number of initial nodes (default is 1)
    m : int
        Number of connections at each step (default is 1)
        m can never be larger than m0.
    seed : int
        Seed for the random number generator (for reproducible graphs).

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> G = graphs.BarabasiAlbert(N=150, seed=42)
    >>> G.set_coordinates(kind='spring', seed=42)
    >>> fig, axes = plt.subplots(1, 2)
    >>> _ = axes[0].spy(G.W, markersize=2)
    >>> _ = G.plot(ax=axes[1])

    """
    def __init__(self, N=1000, m0=1, m=1, seed=None, **kwargs):

        if m > m0:
            raise ValueError('Parameter m cannot be above parameter m0.')

        self.m0 = m0
        self.m = m
        self.seed = seed

        W = sparse.lil_matrix((N, N))
        rng = np.random.default_rng(seed)

        for i in range(m0, N):
            distr = W.sum(axis=1)
            distr += np.concatenate((np.ones((i, 1)), np.zeros((N-i, 1))))

            connections = rng.choice(
                N, size=m, replace=False, p=np.ravel(distr / distr.sum()))
            for elem in connections:
                W[elem, i] = 1
                W[i, elem] = 1

        super(BarabasiAlbert, self).__init__(W, **kwargs)

    def _get_extra_repr(self):
        return dict(m0=self.m0, m=self.m, seed=self.seed)



r"""
Fourier basis
=============

The eigenvectors of the graph Laplacian form the Fourier basis.
The eigenvalues are a measure of variation of their corresponding eigenvector.
The lower the eigenvalue, the smoother the eigenvector. They are hence a
measure of "frequency".

In classical signal processing, Fourier modes are completely delocalized, like
on the grid graph. For general graphs however, Fourier modes might be
localized. See :attr:`pygsp2.graphs.Graph.coherence`.
"""

import numpy as np
from matplotlib import pyplot as plt
import pygsp2 as pg

n_eigenvectors = 7

fig, axes = plt.subplots(2, 7, figsize=(15, 4))

def plot_eigenvectors(G, axes):
    G.compute_fourier_basis(n_eigenvectors)
    limits = [f(G.U) for f in (np.min, np.max)]
    for i, ax in enumerate(axes):
        G.plot(G.U[:, i], limits=limits, colorbar=False, vertex_size=50, ax=ax)
        energy = abs(G.dirichlet_energy(G.U[:, i]))
        ax.set_title(r'$u_{0}^\top L u_{0} = {1:.2f}$'.format(i+1, energy))
        ax.set_axis_off()

G = pg.graphs.Grid2d(10, 10)
plot_eigenvectors(G, axes[0])
fig.subplots_adjust(hspace=0.5, right=0.8)
cax = fig.add_axes([0.82, 0.60, 0.01, 0.26])
fig.colorbar(axes[0, -1].collections[1], cax=cax, ticks=[-0.2, 0, 0.2])

G = pg.graphs.Sensor(seed=42)
plot_eigenvectors(G, axes[1])
fig.subplots_adjust(hspace=0.5, right=0.8)
cax = fig.add_axes([0.82, 0.16, 0.01, 0.26])
_ = fig.colorbar(axes[1, -1].collections[1], cax=cax, ticks=[-0.4, 0, 0.4])
plt.show()


# -*- coding: utf-8 -*-

import numpy as np

from pygsp2.graphs import NNGraph  # prevent circular import in Python < 3.5


class Cube(NNGraph):
    r"""Hyper-cube (NN-graph).

    Parameters
    ----------
    radius : float
        Edge lenght (default = 1)
    nb_pts : int
        Number of vertices (default = 300)
    nb_dim : int
        Dimension (default = 3)
    sampling : string
        Variance of the distance kernel (default = 'random')
        (Can now only be 'random')
    seed : int
        Seed for the random number generator (for reproducible graphs).

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> G = graphs.Cube(seed=42)
    >>> fig = plt.figure()
    >>> ax1 = fig.add_subplot(121)
    >>> ax2 = fig.add_subplot(122, projection='3d')
    >>> _ = ax1.spy(G.W, markersize=0.5)
    >>> _ = G.plot(ax=ax2)

    """

    def __init__(self,
                 radius=1,
                 nb_pts=300,
                 nb_dim=3,
                 sampling='random',
                 seed=None,
                 **kwargs):

        self.radius = radius
        self.nb_pts = nb_pts
        self.nb_dim = nb_dim
        self.sampling = sampling
        self.seed = seed
        rs = np.random.RandomState(seed)

        if self.nb_dim > 3:
            raise NotImplementedError("Dimension > 3 not supported yet!")

        if self.sampling == "random":
            if self.nb_dim == 2:
                pts = rs.rand(self.nb_pts, self.nb_dim)

            elif self.nb_dim == 3:
                n = self.nb_pts // 6

                pts = np.zeros((n*6, 3))
                pts[:n, 1:] = rs.rand(n, 2)
                pts[n:2*n, :] = np.concatenate((np.ones((n, 1)),
                                                rs.rand(n, 2)),
                                               axis=1)

                pts[2*n:3*n, :] = np.concatenate((rs.rand(n, 1),
                                                  np.zeros((n, 1)),
                                                  rs.rand(n, 1)),
                                                 axis=1)
                pts[3*n:4*n, :] = np.concatenate((rs.rand(n, 1),
                                                  np.ones((n, 1)),
                                                  rs.rand(n, 1)),
                                                 axis=1)

                pts[4*n:5*n, :2] = rs.rand(n, 2)
                pts[5*n:6*n, :] = np.concatenate((rs.rand(n, 2),
                                                  np.ones((n, 1))),
                                                 axis=1)

        else:
            raise ValueError("Unknown sampling !")

        plotting = {
            'vertex_size': 80,
            'elevation': 15,
            'azimuth': 0,
            'distance': 9,
        }

        super(Cube, self).__init__(Xin=pts, k=10,
                                   center=False, rescale=False,
                                   plotting=plotting, **kwargs)

    def _get_extra_repr(self):
        attrs = {'radius': '{:.2f}'.format(self.radius),
                 'nb_pts': self.nb_pts,
                 'nb_dim': self.nb_dim,
                 'sampling': self.sampling,
                 'seed': self.seed}
        attrs.update(super(Cube, self)._get_extra_repr())
        return attrs



# -*- coding: utf-8 -*-

import numpy as np

from . import Graph  # prevent circular import in Python < 3.5


class FullConnected(Graph):
    r"""Fully connected graph.

    All weights are set to 1. There is no self-connections.

    Parameters
    ----------
    N : int
        Number of vertices (default = 10)

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> G = graphs.FullConnected(N=20)
    >>> G.set_coordinates(kind='spring', seed=42)
    >>> fig, axes = plt.subplots(1, 2)
    >>> _ = axes[0].spy(G.W, markersize=5)
    >>> _ = G.plot(ax=axes[1])

    """

    def __init__(self, N=10, **kwargs):

        W = np.ones((N, N)) - np.identity(N)
        plotting = {'limits': np.array([-1, 1, -1, 1])}

        super(FullConnected, self).__init__(W, plotting=plotting, **kwargs)



# -*- coding: utf-8 -*-

import numpy as np
from scipy import sparse

from pygsp2 import utils
from . import Graph  # prevent circular import in Python < 3.5


class StochasticBlockModel(Graph):
    r"""Stochastic Block Model (SBM).

    The Stochastic Block Model graph is constructed by connecting nodes with a
    probability which depends on the cluster of the two nodes.  One can define
    the clustering association of each node, denoted by vector z, but also the
    probability matrix M.  All edge weights are equal to 1. By default, Mii >
    Mjk and nodes are uniformly clusterized.

    Parameters
    ----------
    N : int
        Number of nodes (default is 1024).
    k : float
        Number of classes (default is 5).
    z : array_like
        the vector of length N containing the association between nodes and
        classes (default is random uniform).
    M : array_like
        the k by k matrix containing the probability of connecting nodes based
        on their class belonging (default using p and q).
    p : float or array_like
        the diagonal value(s) for the matrix M. If scalar they all have the
        same value. Otherwise expect a length k vector (default is p = 0.7).
    q : float or array_like
        the off-diagonal value(s) for the matrix M. If scalar they all have the
        same value. Otherwise expect a k x k matrix, diagonal will be
        discarded (default is q = 0.3/k).
    directed : bool
        Allow directed edges if True (default is False).
    self_loops : bool
        Allow self loops if True (default is False).
    connected : bool
        Force the graph to be connected (default is False).
    n_try : int or None
        Maximum number of trials to get a connected graph. If None, it will try
        forever.
    seed : int
        Seed for the random number generator (for reproducible graphs).

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> G = graphs.StochasticBlockModel(
    ...     100, k=3, p=[0.4, 0.6, 0.3], q=0.02, seed=42)
    >>> G.set_coordinates(kind='spring', seed=42)
    >>> fig, axes = plt.subplots(1, 2)
    >>> _ = axes[0].spy(G.W, markersize=0.8)
    >>> _ = G.plot(ax=axes[1])

    """

    def __init__(self, N=1024, k=5, z=None, M=None, p=0.7, q=None,
                 directed=False, self_loops=False, connected=False,
                 n_try=10, seed=None, **kwargs):

        self.k = k
        self.directed = directed
        self.self_loops = self_loops
        self.connected = connected
        self.n_try = n_try
        self.seed = seed

        rng = np.random.default_rng(seed)

        if z is None:
            z = rng.integers(0, k, N)
            z.sort()  # Sort for nice spy plot of W, where blocks are apparent.
        self.z = z

        if M is None:

            self.p = p
            p = np.asanyarray(p)
            if p.size == 1:
                p = p * np.ones(k)
            if p.shape != (k,):
                raise ValueError('Optional parameter p is neither a scalar '
                                 'nor a vector of length k.')

            if q is None:
                q = 0.3 / k
            self.q = q
            q = np.asanyarray(q)
            if q.size == 1:
                q = q * np.ones((k, k))
            if q.shape != (k, k):
                raise ValueError('Optional parameter q is neither a scalar '
                                 'nor a matrix of size k x k.')

            M = q
            M.flat[::k+1] = p  # edit the diagonal terms

        self.M = M

        if (M < 0).any() or (M > 1).any():
            raise ValueError('Probabilities should be in [0, 1].')

        # TODO: higher memory, lesser computation alternative.
        # Along the lines of np.random.uniform(size=(N, N)) < p.
        # Or similar to sparse.random(N, N, p, data_rvs=lambda n: np.ones(n)).

        while (n_try is None) or (n_try > 0):

            nb_row, nb_col = 0, 0
            csr_data, csr_i, csr_j = [], [], []
            for _ in range(N**2):
                if nb_row != nb_col or self_loops:
                    if nb_row >= nb_col or directed:
                        if rng.uniform() < M[z[nb_row], z[nb_col]]:
                            csr_data.append(1)
                            csr_i.append(nb_row)
                            csr_j.append(nb_col)
                if nb_row < N-1:
                    nb_row += 1
                else:
                    nb_row = 0
                    nb_col += 1

            W = sparse.csr_matrix((csr_data, (csr_i, csr_j)), shape=(N, N))

            if not directed:
                W = utils.symmetrize(W, method='tril')

            if not connected:
                break
            if Graph(W).is_connected():
                break
            if n_try is not None:
                n_try -= 1
        if connected and n_try == 0:
            raise ValueError('The graph could not be connected after {} '
                             'trials. Increase the connection probability '
                             'or the number of trials.'.format(self.n_try))

        self.info = {'node_com': z, 'comm_sizes': np.bincount(z),
                     'world_rad': np.sqrt(N)}

        super(StochasticBlockModel, self).__init__(W, **kwargs)

    def _get_extra_repr(self):
        attrs = {'k': self.k}
        if type(self.p) is float:
            attrs['p'] = '{:.2f}'.format(self.p)
        if type(self.q) is float:
            attrs['q'] = '{:.2f}'.format(self.q)
        attrs.update({'directed': self.directed,
                      'self_loops': self.self_loops,
                      'connected': self.connected,
                      'seed': self.seed})
        return attrs



# -*- coding: utf-8 -*-

"""
Test suite for the learning module of the pygsp2 package.

"""

import unittest

import numpy as np

from pygsp2 import graphs, filters, learning
from scipy import spatial


class TestCase(unittest.TestCase):

    def test_regression_tikhonov_1(self):
        """Solve a trivial regression problem."""
        G = graphs.Ring(N=8)
        signal = np.array([0, np.nan, 4, np.nan, 4, np.nan, np.nan, np.nan])
        signal_bak = signal.copy()
        mask = np.array([True, False, True, False, True, False, False, False])
        truth = np.array([0, 2, 4, 4, 4, 3, 2, 1])
        recovery = learning.regression_tikhonov(G, signal, mask, tau=0)
        np.testing.assert_allclose(recovery, truth)

        # Test the numpy solution.
        G = graphs.Graph(G.W.toarray())
        recovery = learning.regression_tikhonov(G, signal, mask, tau=0)
        np.testing.assert_allclose(recovery, truth)
        np.testing.assert_allclose(signal_bak, signal)

    def test_regression_tikhonov_2(self):
        """Solve a regression problem with a constraint."""
        G = graphs.Sensor(100)
        G.estimate_lmax()

        # Create a smooth signal.
        filt = filters.Filter(G, lambda x: 1 / (1 + 10*x))
        rng = np.random.default_rng(1)
        signal = filt.analyze(rng.normal(size=(G.n_vertices, 5)))

        # Make the input signal.
        mask = rng.uniform(0, 1, G.n_vertices) > 0.5
        measures = signal.copy()
        measures[~mask] = np.nan
        measures_bak = measures.copy()

        # Solve the problem.
        recovery0 = learning.regression_tikhonov(G, measures, mask, tau=0)
        np.testing.assert_allclose(measures_bak, measures)

        recovery1 = np.zeros_like(recovery0)
        for i in range(recovery0.shape[1]):
            recovery1[:, i] = learning.regression_tikhonov(
                G, measures[:, i], mask, tau=0)
        np.testing.assert_allclose(measures_bak, measures)

        G = graphs.Graph(G.W.toarray())
        recovery2 = learning.regression_tikhonov(G, measures, mask, tau=0)
        recovery3 = np.zeros_like(recovery0)
        for i in range(recovery0.shape[1]):
            recovery3[:, i] = learning.regression_tikhonov(
                G, measures[:, i], mask, tau=0)

        np.testing.assert_allclose(recovery1, recovery0)
        np.testing.assert_allclose(recovery2, recovery0)
        np.testing.assert_allclose(recovery3, recovery0)
        np.testing.assert_allclose(measures_bak, measures)

    def test_regression_tikhonov_3(self, tau=3.5):
        """Solve a relaxed regression problem."""
        G = graphs.Sensor(100)
        G.estimate_lmax()

        # Create a smooth signal.
        filt = filters.Filter(G, lambda x: 1 / (1 + 10*x))
        rng = np.random.default_rng(1)
        signal = filt.analyze(rng.normal(size=(G.n_vertices, 6)))

        # Make the input signal.
        mask = rng.uniform(0, 1, G.n_vertices) > 0.5
        measures = signal.copy()
        measures[~mask] = 18
        measures_bak = measures.copy()

        L = G.L.toarray()
        recovery = np.matmul(np.linalg.inv(np.diag(1*mask) + tau * L),
                             (mask * measures.T).T)

        # Solve the problem.
        recovery0 = learning.regression_tikhonov(G, measures, mask, tau=tau)
        np.testing.assert_allclose(measures_bak, measures)
        recovery1 = np.zeros_like(recovery0)
        for i in range(recovery0.shape[1]):
            recovery1[:, i] = learning.regression_tikhonov(
                G, measures[:, i], mask, tau)
        np.testing.assert_allclose(measures_bak, measures)

        G = graphs.Graph(G.W.toarray())
        recovery2 = learning.regression_tikhonov(G, measures, mask, tau)
        recovery3 = np.zeros_like(recovery0)
        for i in range(recovery0.shape[1]):
            recovery3[:, i] = learning.regression_tikhonov(
                G, measures[:, i], mask, tau)

        np.testing.assert_allclose(recovery0, recovery, atol=1e-5)
        np.testing.assert_allclose(recovery1, recovery, atol=1e-5)
        np.testing.assert_allclose(recovery2, recovery, atol=1e-5)
        np.testing.assert_allclose(recovery3, recovery, atol=1e-5)
        np.testing.assert_allclose(measures_bak, measures)

    def test_classification_tikhonov(self):
        """Solve a classification problem."""
        G = graphs.Logo()
        signal = np.zeros([G.n_vertices], dtype=int)
        signal[G.info['idx_s']] = 1
        signal[G.info['idx_p']] = 2

        # Make the input signal.
        rng = np.random.default_rng(2)
        mask = rng.uniform(size=G.n_vertices) > 0.3

        measures = signal.copy()
        measures[~mask] = -1
        measures_bak = measures.copy()

        # Solve the classification problem.
        recovery = learning.classification_tikhonov(G, measures, mask, tau=0)
        recovery = np.argmax(recovery, axis=1)

        np.testing.assert_array_equal(recovery, signal)

        # Test the function with the simplex projection.
        recovery = learning.classification_tikhonov_simplex(
            G, measures, mask, tau=0.1)

        # Assert that the probabilities sums to 1
        np.testing.assert_allclose(np.sum(recovery, axis=1), 1)

        # Check the quality of the solution.
        recovery = np.argmax(recovery, axis=1)
        np.testing.assert_allclose(signal, recovery)
        np.testing.assert_allclose(measures_bak, measures)

    def test_graph_log_degree(self):
        "Solve graph learning from signal"

        # Make graph
        G = graphs.Graph([[0, 0, 0, 0],
                          [0, 0, 1, 1],
                          [0, 1, 0, 1],
                          [0, 1, 1, 0]])

        # Signal
        signal = np.array([0, 1, 1, 1])

        # Compute the distance matrix
        kdt = spatial.KDTree(signal[:, None])

        D, NN = kdt.query(signal[:, None], k=len(signal))

        Z = np.zeros((G.N, G.N))

        for i, n in enumerate(NN):
            Z[i, n] = D[i]

        # Learn graph
        A = 0.2
        B = 0.01

        W = learning.graph_log_degree(Z, A, B, w_max=1)
        W[W < 1e-1] = 0

        np.testing.assert_allclose(W, G.W.toarray())


suite = unittest.TestLoader().loadTestsFromTestCase(TestCase)



# -*- coding: utf-8 -*-

r"""
The :mod:`pygsp2` package is mainly organized around the following two modules:

* :mod:`.graphs` to create and manipulate various kinds of graphs,
* :mod:`.filters` to create and manipulate various graph filters.

Moreover, the following modules provide additional functionality:

* :mod:`.plotting` to plot,
* :mod:`.reduction` to reduce a graph while keeping its structure,
* :mod:`.features` to compute features on graphs,
* :mod:`.learning` to solve learning problems,
* :mod:`.optimization` to help solving convex optimization problems,
* :mod:`.utils` for various utilities.
* :mod:`.utils_examples` for metro applications.

"""

from . import graphs  # noqa: F401
from . import filters  # noqa: F401
from . import plotting  # noqa: F401
from . import reduction  # noqa: F401
from . import features  # noqa: F401
from . import learning  # noqa: F401
from . import optimization  # noqa: F401
from . import utils  # noqa: F401
from . import utils_examples # noqa: F401

# Users only call the plot methods from the objects.
# It's thus more convenient for them to have the doc there.
# But it's more convenient for developers to have the doc alongside the code.
try:
    filters.Filter.plot.__doc__ = plotting._plot_filter.__doc__
    graphs.Graph.plot.__doc__ = plotting._plot_graph.__doc__
    graphs.Graph.plot_spectrogram.__doc__ = plotting._plot_spectrogram.__doc__
except AttributeError:
    # For Python 2.7.
    filters.Filter.plot.__func__.__doc__ = plotting._plot_filter.__doc__
    graphs.Graph.plot.__func__.__doc__ = plotting._plot_graph.__doc__
    graphs.Graph.plot_spectrogram.__func__.__doc__ = plotting._plot_spectrogram.__doc__

__version__ = '0.5.1'
__release_date__ = '2017-12-15'


def test():  # pragma: no cover
    """Run the test suite."""
    import unittest
    # Lazy as it might be slow and require additional dependencies.
    from pygsp2.tests import suite
    unittest.TextTestRunner(verbosity=2).run(suite)



# -*- coding: utf-8 -*-

"""
Test suite for the utils module of the pygsp2 package.

"""

import unittest

import numpy as np
from scipy import sparse

from pygsp2 import graphs, utils


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def test_symmetrize(self):
        W = sparse.random(100, 100, random_state=42)
        for method in ['average', 'maximum', 'fill', 'tril', 'triu']:
            # Test that the regular and sparse versions give the same result.
            W1 = utils.symmetrize(W, method=method)
            W2 = utils.symmetrize(W.toarray(), method=method)
            np.testing.assert_equal(W1.toarray(), W2)
        self.assertRaises(ValueError, utils.symmetrize, W, 'sum')


suite = unittest.TestLoader().loadTestsFromTestCase(TestCase)



r"""
Random walks
============

Probability of a random walker to be on any given vertex after a given number
of steps starting from a given distribution.
"""

# sphinx_gallery_thumbnail_number = 2

import numpy as np
from scipy import sparse
from matplotlib import pyplot as plt
import pygsp2 as pg

N = 7
steps = [0, 1, 2, 3]

graph = pg.graphs.Grid2d(N)
delta = np.zeros(graph.N)
delta[N//2*N + N//2] = 1

probability = sparse.diags(graph.dw**(-1)).dot(graph.W)

fig, axes = plt.subplots(1, len(steps), figsize=(12, 3))
for step, ax in zip(steps, axes):
    state = (probability**step).__rmatmul__(delta)  ## = delta @ probability**step
    graph.plot(state, ax=ax, title=r'$\delta P^{}$'.format(step))
    ax.set_axis_off()

fig.tight_layout()

###############################################################################
# Stationary distribution.

graphs = [
    pg.graphs.Ring(10),
    pg.graphs.Grid2d(5),
    pg.graphs.Comet(8, 4),
    pg.graphs.BarabasiAlbert(20, seed=42),
]

fig, axes = plt.subplots(1, len(graphs), figsize=(12, 3))

for graph, ax in zip(graphs, axes):

    if not hasattr(graph, 'coords'):
        graph.set_coordinates(seed=10)

    P = sparse.diags(graph.dw**(-1)).dot(graph.W)

#    e, u = np.linalg.eig(P.T.toarray())
#    np.testing.assert_allclose(np.linalg.inv(u.T) @ np.diag(e) @ u.T,
#                               P.toarray(), atol=1e-10)
#    np.testing.assert_allclose(np.abs(e[0]), 1)
#    stationary = np.abs(u.T[0])

    e, u = sparse.linalg.eigs(P.T, k=1, which='LR')
    np.testing.assert_allclose(e, 1)
    stationary = np.abs(u).squeeze()
    assert np.all(stationary < 0.71)

    colorbar = False if type(graph) is pg.graphs.Ring else True
    graph.plot(stationary, colorbar=colorbar, ax=ax, title='$xP = x$')
    ax.set_axis_off()

fig.tight_layout()
plt.show()



# -*- coding: utf-8 -*-

from __future__ import division

import numpy as np

from . import Filter  # prevent circular import in Python < 3.5


class HalfCosine(Filter):
    r"""Design an half cosine filter bank (tight frame).

    Parameters
    ----------
    G : graph
    Nf : int
        Number of filters from 0 to lmax (default = 6)

    Examples
    --------

    Filter bank's representation in Fourier and time (ring graph) domains.

    >>> import matplotlib.pyplot as plt
    >>> G = graphs.Ring(N=20)
    >>> G.estimate_lmax()
    >>> G.set_coordinates('line1D')
    >>> g = filters.HalfCosine(G)
    >>> s = g.localize(G.N // 2)
    >>> fig, axes = plt.subplots(1, 2)
    >>> _ = g.plot(ax=axes[0])
    >>> _ = G.plot(s, ax=axes[1])

    """

    def __init__(self, G, Nf=6):

        if Nf <= 2:
            raise ValueError('The number of filters must be greater than 2.')

        dila_fact = G.lmax * 3 / (Nf - 2)

        def kernel(x):
            y = np.cos(2 * np.pi * (x / dila_fact - .5))
            y = np.multiply((.5 + .5*y), (x >= 0))
            return np.multiply(y, (x <= dila_fact))

        kernels = []

        for i in range(Nf):

            def kernel_centered(x, i=i):
                return kernel(x - dila_fact/3 * (i - 2))

            kernels.append(kernel_centered)

        super(HalfCosine, self).__init__(G, kernels)



# -*- coding: utf-8 -*-

import numpy as np
from scipy import sparse

from . import Graph  # prevent circular import in Python < 3.5


class LowStretchTree(Graph):
    r"""Low stretch tree.

    Build the root of a low stretch tree on a grid of points. There are
    :math:`2k` points on each side of the grid, and therefore :math:`2^{2k}`
    vertices in total. The edge weights are all equal to 1.

    Parameters
    ----------
    k : int
        :math:`2^k` points on each side of the grid of vertices.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> G = graphs.LowStretchTree(k=2)
    >>> fig, axes = plt.subplots(1, 2)
    >>> _ = axes[0].spy(G.W)
    >>> _ = G.plot(ax=axes[1])

    """

    def __init__(self, k=6, **kwargs):

        self.k = k

        XCoords = np.array([1, 2, 1, 2], dtype=int)
        YCoords = np.array([1, 1, 2, 2], dtype=int)

        ii = np.array([0, 0, 1, 2, 2, 3], dtype=int)
        jj = np.array([1, 2, 1, 3, 0, 2], dtype=int)

        for p in range(1, k):
            ii = np.concatenate((ii, ii + 4**p, ii + 2*4**p, ii + 3*4**p,
                                 [4**p - 1], [4**p - 1],
                                 [4**p + (4**(p+1) + 2) // 3 - 1],
                                 [(5*4**p + 1) // 3 - 1],
                                 [4**p + (4**(p+1) + 2) // 3 - 1], [3*4**p]))
            jj = np.concatenate((jj, jj + 4**p, jj + 2*4**p, jj + 3*4**p,
                                 [(5*4**p + 1) // 3 - 1],
                                 [4**p + (4**(p+1) + 2) // 3 - 1],
                                 [3*4**p], [4**p - 1], [4**p - 1],
                                 [4**p + (4**(p+1) + 2) // 3 - 1]))

            YCoords = np.kron(np.ones((2), dtype=int), YCoords)
            YCoords = np.concatenate((YCoords, YCoords + 2**p))

            XCoords = np.concatenate((XCoords, XCoords + 2**p))
            XCoords = np.kron(np.ones((2), dtype=int), XCoords)

        W = sparse.csc_matrix((np.ones_like(ii), (ii, jj)))
        coords = np.concatenate((XCoords[:, np.newaxis],
                                 YCoords[:, np.newaxis]),
                                axis=1)

        self.root = 4**(k - 1)

        plotting = {"edges_width": 1.25,
                    "vertex_size": 75,
                    "limits": np.array([0, 2**k + 1, 0, 2**k + 1])}

        super(LowStretchTree, self).__init__(W,
                                             coords=coords,
                                             plotting=plotting,
                                             **kwargs)

    def _get_extra_repr(self):
        return dict(k=self.k)



r"""
Utils Examples
=====

Utils functions used in the examples.

This module contains functions to run examples of graph signal processing.

Functions are made to avoid excesive repetition of loading, plotting and
database management. Bleow there is a list of the examples that use these
functions:

* metro_graph_signal.py
* metro_graph_regression.py
* metro_simulation.py
* metro_simulation2.py
"""
import os
import json
import utm
import networkx as nx
from geopy.distance import distance
import matplotlib
import matplotlib.pyplot as plt
from matplotlib import colormaps
import numpy as np
import zipfile
import io
import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

ASSETS_METRO = [
    {
        "filename": "2023.11 Matriz_baj_SS_MH.xlsb",
        "url": r"https://zenodo.org/records/12822782/files/tablas.zip"
    },
    {
        "filename": "santiago_metro_stations_coords.geojson",
        "url": r"https://zenodo.org/records/11637462/files/santiago_metro_stations_coords.geojson"
    },
    {
        "filename": "santiago_metro_stations_connections.txt",
        "url": r"https://zenodo.org/records/11637462/files/santiago_metro_stations_connections.txt"
    }
]

ASSETS = {
    "metro": ASSETS_METRO
}

def make_metro_graph(edgesfile='santiago_metro_stations_connections.txt',
                     coordsfile='santiago_metro_stations_coords.geojson'):
    """Create a NetworkX graph corresponding to Santiago Metro network.

    Parameters
    ----------
    edgesfile : str, optional.
        File name that contains the edges of the connections.  See notes for
        the link to download the file.

    coordsfile : str, optional.
        Name of the file with the coordenates of the nodes for the graph.  See
        notes for the link to download the file.

    Returns
    -------
    G : Network Graph.
        Graph data structure containing the graph and its attributes.
    pos : dict.
        Contains the coordenates of each station. Keys are the station names.
        Names can be obtained with `list(G)`.

    Notes
    -----
    * Download the file `santiago_metro_stations_coords.geojson` from this link:

    https://zenodo.org/records/11637462/files/santiago_metro_stations_coords.geojson

    * Download the file `santiago_metro_stations_connections.txt` from this link:

    https://zenodo.org/records/11637462/files/santiago_metro_stations_connections.txt
    """

    with open(coordsfile) as f:
        data = json.load(f)

    d = {x['properties']['name']: x['geometry']['coordinates'] for
         x in data['features']}

    # Manually add missing stations
    d['Irarrázaval'] = (-70.6283197, -33.4550371)
    d['Vicente Valdés'] = (-70.5967924, -33.5264186)

    G = nx.Graph()
    ref_station = 'San Pablo'
    for station, latlong in d.items():
        x, y, _, _ = utm.from_latlon(*latlong)
        G.add_node(station, latlong=latlong,
                   distance=distance(latlong, d[ref_station]).meters,
                   x=x, y=y)

    with open(edgesfile) as f:
        for e in f.readlines():
            if e[0] == '#' or len(e) < 2:
                continue
            u, v = e.split(',')
            G.add_edge(u.strip(), v.strip())

    pos = {node: (G.nodes[node]['y'], G.nodes[node]['x']) for node in G.nodes}

    return G, pos


def metro_database_preprocessing(commutes, stations):
    """Preprocess commute database.

    Preprocess commute dataframe to match with the metro network names in
    `santiago_metro_stations_connections.txt`.

    Parameters
    ----------
    commutes : pandas DataFrame.
        The `DataFrame` is loaded from https://www.dtpm.cl/index.php/documentos/matrices-de-viaje
        the files in "tablas de Subidas y Bajadas" have to be downloaded.
    stations : list of strings.
        `stations` should contain the name of metro stations
        in the graph ordered by the node index. In a networkx
        graph this can be obtained through `list(G)`.

    Returns
    -------
    metro_commutes : pandas DataFrame.
        Rows in the given database that correspond to metro commutes.
    stations : numpy array.
        One dimensional array with the values in the `TOTAL` column. The length
        of the array is the number of stations.
    """
    # %% Use only metro commutes in database
    idx = ['L' in servicio for servicio in commutes['servicio Sonda']]
    metro_commutes = commutes[idx]

    # Change some names to coincide with node names
    metro_commutes.loc[metro_commutes['paradero'] == 'CAL Y CANTO',
                       'paradero'] = 'PUENTE CAL Y CANTO'
    metro_commutes.loc[metro_commutes['paradero'] == 'PARQUE OHIGGINS',
                       'paradero'] = 'PARQUE O\'HIGGINS'
    metro_commutes.loc[metro_commutes['paradero'] == 'PDTE PEDRO AGUIRRE CERDA',
                       'paradero'] = 'PRESIDENTE PEDRO AGUIRRE CERDA'
    metro_commutes.loc[metro_commutes['paradero'] == 'PLAZA MAIPU',
                       'paradero'] = 'PLAZA DE MAIPU'
    metro_commutes.loc[metro_commutes['paradero'] == 'RONDIZONNI',
                       'paradero'] = 'RONDIZZONI'
    metro_commutes.loc[metro_commutes['paradero'] == 'UNION LATINO AMERICANA',
                       'paradero'] = 'UNION LATINOAMERICANA'

    # %% Build signal bassed on the stations
    signal = np.zeros_like(stations, dtype=float)

    for value, station in zip(metro_commutes['TOTAL'], metro_commutes['paradero']):
        graph_idx = [station == station_graph for station_graph in stations]
        signal[graph_idx] = float(value)

    return metro_commutes, signal


def plot_signal_in_graph(G, signal, title='Graph Signal', label='', cmap='viridis', alpha=1):
    """Function to plot signal in graph using networkx.

    Parameters
    ----------
    G : networkx Graph.
    signal: numpy array.
        Vector with signal values for each node. Must have the same length as
        number of nodes in `G`.
    title : str, optional.
        Title in the graph.
    label : str, optional.
        Lables to be displayed in colorbar.
    cmap : str, optional.
        Sets colormap
    alpha : float, optional.
        Sets transparency

    Returns
    -------
    fig : matplotlib figure.
    ax : matplotlib.pyplot axes.
    """
    pos = {node: (G.nodes[node]['y'], G.nodes[node]['x']) for node in G.nodes}
    # Map signal to a color
    if cmap in colormaps:
        cmap = matplotlib.colormaps.get_cmap(cmap)
    else:
        print("Wrong colormap")
        cmap = matplotlib.colormaps.get_cmap('viridis')

    alpha = np.abs(alpha)

    normalized_signal = signal / np.max(signal)
    colors = cmap(normalized_signal)

    # Initialize figure
    fig, ax = plt.subplots(figsize=(10, 7))
    # Draw edges and nodes
    nx.draw_networkx_edges(G, pos, node_size=20, ax=ax)
    pc = nx.draw_networkx_nodes(G, pos, node_color=colors, node_size=20, ax=ax, alpha=alpha)
    cbar = plt.colorbar(pc, ticks=[0, 0.5, 1], label=label)
    cbar.set_ticklabels([f'{label:.0f}' for label in [
                        0, np.amax(signal)/2, np.amax(signal)]])
    plt.title(title)
    plt.tight_layout()
    plt.set_cmap(cmap) 

    return fig, ax

def fetch_data(output_dir, database="metro"):
    """ 
    Fetch data from the internet and save it in the output_dir.

    Parameters
    ----------
    output_dir : str
        Directory where the data will be saved.
    database : str, optional
        Database to fetch data from. Options are: "metro".
    """ 
    for asset_dict in ASSETS[database]:
        filename = asset_dict["filename"]
        url = asset_dict["url"]
        assets_filepath = os.path.join(output_dir, filename)
        if not os.path.isfile(assets_filepath):
            print(f'Downloading data file to:\n {assets_filepath}')
            r = get_with_retries(url) #requests.get(url, timeout=30)
            if url.endswith('.zip'):
                z = zipfile.ZipFile(io.BytesIO(r.content))
                z.extractall(output_dir)
            else:
                with open(assets_filepath, 'wb') as f:
                    f.write(r.content)


def get_with_retries(url, retries=3, backoff_factor=0.3, status_forcelist=(500, 502, 504), timeout=10):
    # Create a session
    session = requests.Session()
    
    # Define the retry strategy
    retry_strategy = Retry(
        total=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
        allowed_methods=["HEAD", "GET", "OPTIONS"]
    )
    
    # Mount the adapter with the retry strategy
    adapter = HTTPAdapter(max_retries=retry_strategy)
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    
    try:
        # Send a GET request with timeout
        response = session.get(url, timeout=timeout)
        
        # Check if the request was successful
        if response.status_code == 200:
            # Process the response data
            print("File downloaded successfully.")
            return response
        else:
            print(f"Request failed with status code: {response.status_code}")
    
    except requests.exceptions.ConnectTimeout:
        print("The request timed out while trying to connect to the remote server.")
    
    except requests.exceptions.ReadTimeout:
        print("The server did not send any data in the allotted amount of time.")
    
    except requests.exceptions.RequestException as e:
        # Handle any exceptions that occur
        print(f"An error occurred: {e}")



# -*- coding: utf-8 -*-

from __future__ import division

import numpy as np

from . import Filter  # prevent circular import in Python < 3.5


class Rectangular(Filter):
    r"""Design a rectangular filter.

    The filter evaluates at one in the interval [band_min, band_max] and zero
    everywhere else.

    The rectangular kernel is defined as

    .. math:: g(\lambda) = \begin{cases}
        0 & \text{if } \lambda < \text{band}_\text{min}, \\
        1 & \text{if band}_\text{min} \leq \lambda \leq \text{band}_\text{max}, \\
        0 & \text{if } \lambda > \text{band}_\text{max}, \\
                            \end{cases}

    where :math:`\lambda \in [0, 1]` corresponds to the normalized graph
    eigenvalues.

    Parameters
    ----------
    G : graph
    band_min : float
        Minimum relative band. The filter evaluates at 1 at this frequency.
        Zero corresponds to the smallest eigenvalue (which is itself equal to
        zero), one corresponds to the largest eigenvalue.
        If None, the filter has no lower bound (which corresponds to
        :math:`\text{band}_\text{min} = -\infty`) and is high-pass.
    band_max : float
        Maximum relative band. The filter evaluates at 1 at this frequency.
        If None, the filter has no upper bound (which corresponds to
        :math:`\text{band}_\text{min} = \infty`) and is high-pass.

    Examples
    --------

    Filter bank's representation in Fourier and time (ring graph) domains.

    >>> import matplotlib.pyplot as plt
    >>> G = graphs.Ring(N=20)
    >>> G.estimate_lmax()
    >>> G.set_coordinates('line1D')
    >>> g = filters.Rectangular(G, band_min=0.1, band_max=0.5)
    >>> s = g.localize(G.N // 2)
    >>> fig, axes = plt.subplots(1, 2)
    >>> _ = g.plot(ax=axes[0])
    >>> _ = G.plot(s, ax=axes[1])

    """

    def __init__(self, G, band_min=None, band_max=0.2):

        self.band_min = band_min
        self.band_max = band_max

        def kernel_lowpass(x):
            x = x / G.lmax
            return x <= band_max

        def kernel_highpass(x):
            x = x / G.lmax
            return x >= band_min

        if (band_min is None) and (band_max is None):
            kernel = lambda x: np.ones_like(x)
        elif band_min is None:
            kernel = kernel_lowpass
        elif band_max is None:
            kernel = kernel_highpass
        else:
            kernel = lambda x: kernel_lowpass(x) * kernel_highpass(x)

        super(Rectangular, self).__init__(G, kernel)

    def _get_extra_repr(self):
        attrs = dict()
        if self.band_min is not None:
            attrs.update(band_min='{:.2f}'.format(self.band_min))
        if self.band_max is not None:
            attrs.update(band_max='{:.2f}'.format(self.band_max))
        return attrs



# -*- coding: utf-8 -*-

from __future__ import division

import numpy as np

from . import Filter  # prevent circular import in Python < 3.5


class Heat(Filter):
    r"""Design a filter bank of heat kernels.

    The (low-pass) heat kernel is defined in the spectral domain as

    .. math:: g_\tau(\lambda) = \exp(-\tau \lambda),

    where :math:`\lambda \in [0, 1]` are the normalized eigenvalues of the
    graph Laplacian, and :math:`\tau` is a parameter that captures both time
    and thermal diffusivity.

    The heat kernel is the fundamental solution to the heat equation

    .. math:: - \tau L f(t) = \partial_t f(t),

    where :math:`f: \mathbb{R}_+ \rightarrow \mathbb{R}^N` is the heat
    distribution over the graph at time :math:`t`. Given the initial condition
    :math:`f(0)`, the solution of the heat equation is expressed as

    .. math:: f(t) = e^{-\tau t L} f(0)
                   = U e^{-\tau t \Lambda} U^\top f(0)
                   = g_{\tau t}(L) f(0).

    The above is, by definition, the convolution of the signal :math:`f(0)`
    with the kernel :math:`g_{\tau t}(\lambda) = \exp(-\tau t \lambda)`.
    Hence, applying this filter to a signal simulates heat diffusion.

    Since the kernel is applied to the graph eigenvalues :math:`\lambda`, which
    can be interpreted as squared frequencies, it can also be considered as a
    generalization of the Gaussian kernel on graphs.

    Parameters
    ----------
    G : graph
    scale : float or iterable
        Scaling parameter. When solving heat diffusion, it encompasses both
        time and thermal diffusivity.
        If iterable, creates a filter bank with one filter per value.
    normalize : bool
        Whether to normalize the kernel to have unit L2 norm.
        The normalization needs the eigenvalues of the graph Laplacian.

    Examples
    --------

    Filter bank's representation in Fourier and time (ring graph) domains.

    >>> import matplotlib.pyplot as plt
    >>> G = graphs.Ring(N=20)
    >>> G.estimate_lmax()
    >>> G.set_coordinates('line1D')
    >>> g = filters.Heat(G, scale=[5, 10, 100])
    >>> s = g.localize(G.N // 2)
    >>> fig, axes = plt.subplots(1, 2)
    >>> _ = g.plot(ax=axes[0])
    >>> _ = G.plot(s, ax=axes[1])

    Heat diffusion from two sources on a grid.

    >>> import matplotlib.pyplot as plt
    >>> n_side = 11
    >>> graph = graphs.Grid2d(n_side)
    >>> graph.estimate_lmax()
    >>> sources = [
    ...     (n_side//4 * n_side) + (n_side//4),
    ...     (n_side*3//4 * n_side) + (n_side*3//4),
    ... ]
    >>> delta = np.zeros(graph.n_vertices)
    >>> delta[sources] = 5
    >>> steps = np.array([1, 5])
    >>> diffusivity = 10
    >>> g = filters.Heat(graph, scale=diffusivity*steps)
    >>> diffused = g.filter(delta)
    >>> fig, axes = plt.subplots(1, len(steps), figsize=(10, 4))
    >>> _ = fig.suptitle('Heat diffusion', fontsize=16)
    >>> for i, ax in enumerate(axes):
    ...     _ = graph.plot(diffused[:, i], highlight=sources,
    ...                    title='step {}'.format(steps[i]), ax=ax)
    ...     ax.set_aspect('equal', 'box')
    ...     ax.set_axis_off()

    Normalized heat kernel.

    >>> G = graphs.Logo()
    >>> G.compute_fourier_basis()
    >>> g = filters.Heat(G, scale=5)
    >>> y = g.evaluate(G.e)
    >>> print('norm: {:.2f}'.format(np.linalg.norm(y[0])))
    norm: 9.76
    >>> g = filters.Heat(G, scale=5, normalize=True)
    >>> y = g.evaluate(G.e)
    >>> print('norm: {:.2f}'.format(np.linalg.norm(y[0])))
    norm: 1.00

    """

    def __init__(self, G, scale=10, normalize=False):

        try:
            iter(scale)
        except TypeError:
            scale = [scale]

        self.scale = scale
        self.normalize = normalize

        def kernel(x, scale):
            return np.minimum(np.exp(-scale * x / G.lmax), 1)

        kernels = []
        for s in scale:
            norm = np.linalg.norm(kernel(G.e, s)) if normalize else 1
            kernels.append(lambda x, s=s, norm=norm: kernel(x, s) / norm)

        super(Heat, self).__init__(G, kernels)

    def _get_extra_repr(self):
        scale = '[' + ', '.join('{:.2f}'.format(s) for s in self.scale) + ']'
        return dict(scale=scale, normalize=self.normalize)



# -*- coding: utf-8 -*-

from functools import partial

import numpy as np

from pygsp2 import utils
# prevent circular import in Python < 3.5
from . import approximations
from ..graphs import Graph


_logger = utils.build_logger(__name__)


class Filter(object):
    r"""
    The base Filter class.

    * Provide a common interface (and implementation) to filter objects.
    * Can be instantiated to construct custom filters from functions.
    * Initialize attributes for derived classes.

    Parameters
    ----------
    G : graph
        The graph to which the filter bank is tailored.
    kernels : function or list of functions
        A (list of) function(s) defining the filter bank in the Fourier domain.
        One function per filter.

    Attributes
    ----------
    G : Graph
        The graph to which the filter bank was tailored. It is a reference to
        the graph passed when instantiating the class.
    n_features_in : int
        Number of signals or features the filter bank takes in.
    n_features_out : int
        Number of signals or features the filter bank gives out.
    n_filters : int
        Number of filters in the filter bank.

    Examples
    --------
    >>> G = graphs.Logo()
    >>> my_filter = filters.Filter(G, lambda x: x / (1. + x))
    >>>
    >>> # Signal: Kronecker delta.
    >>> signal = np.zeros(G.N)
    >>> signal[42] = 1
    >>>
    >>> filtered_signal = my_filter.filter(signal)

    """

    def __init__(self, G, kernels):

        self.G = G

        try:
            iter(kernels)
        except TypeError:
            kernels = [kernels]
        self._kernels = kernels

        # Only used by subclasses to instantiate a single filterbank.
        self.n_features_in, self.n_features_out = (1, len(kernels))
        self.shape = (self.n_features_out, self.n_features_in)
        self.n_filters = self.n_features_in * self.n_features_out
        self.Nf = self.n_filters  # TODO: kept for backward compatibility only.

    def _get_extra_repr(self):
        """To be overloaded by children."""
        return dict()

    def __repr__(self):
        attrs = {'in': self.n_features_in, 'out': self.n_features_out}
        attrs.update(self._get_extra_repr())
        s = ''
        for key, value in attrs.items():
            s += '{}={}, '.format(key, value)
        return '{}({})'.format(self.__class__.__name__, s[:-2])

    def __len__(self):
        # Numpy returns shape[0].
        return self.n_filters

    def __getitem__(self, key):
        return Filter(self.G, self._kernels[key])

    def __add__(self, other):
        """Concatenation of filterbanks."""
        if not isinstance(other, Filter):
            return NotImplemented
        return Filter(self.G, self._kernels + other._kernels)

    def __call__(self, x):
        if isinstance(x, Graph):
            return Filter(x, self._kernels)
        else:
            return self.evaluate(x)

    def __matmul__(self, other):
        return self.filter(other)

    def toarray(self):
        r"""Return an array representation of the filter bank.

        See :meth:`compute_frame`.
        """
        return self.compute_frame()

    def evaluate(self, x):
        r"""Evaluate the kernels at given frequencies.

        Parameters
        ----------
        x : array_like
            Graph frequencies at which to evaluate the filter.

        Returns
        -------
        y : ndarray
            Frequency response of the filters. Shape ``(g.Nf, len(x))``.

        Examples
        --------
        Frequency response of a low-pass filter:

        >>> import matplotlib.pyplot as plt
        >>> G = graphs.Logo()
        >>> G.compute_fourier_basis()
        >>> f = filters.Expwin(G)
        >>> G.compute_fourier_basis()
        >>> y = f.evaluate(G.e)
        >>> plt.plot(G.e, y[0])  # doctest: +ELLIPSIS
        [<matplotlib.lines.Line2D object at ...>]

        """
        x = np.asanyarray(x)
        # Avoid to copy data as with np.array([g(x) for g in self._kernels]).
        y = np.empty([self.Nf] + list(x.shape))
        for i, kernel in enumerate(self._kernels):
            y[i] = kernel(x)
        return y

    def filter(self, s, method='chebyshev', order=30):
        r"""Filter signals (analysis or synthesis).

        A signal is defined as a rank-3 tensor of shape ``(N_NODES, N_SIGNALS,
        N_FEATURES)``, where ``N_NODES`` is the number of nodes in the graph,
        ``N_SIGNALS`` is the number of independent signals, and ``N_FEATURES``
        is the number of features which compose a graph signal, or the
        dimensionality of a graph signal. For example if you filter a signal
        with a filter bank of 8 filters, you're extracting 8 features and
        decomposing your signal into 8 parts. That is called analysis. Your are
        thus transforming your signal tensor from ``(G.N, 1, 1)`` to ``(G.N, 1,
        8)``. Now you may want to combine back the features to form an unique
        signal. For this you apply again 8 filters, one filter per feature, and
        sum the result up. As such you're transforming your ``(G.N, 1, 8)``
        tensor signal back to ``(G.N, 1, 1)``. That is known as synthesis. More
        generally, you may want to map a set of features to another, though
        that is not implemented yet.

        The method computes the transform coefficients of a signal :math:`s`,
        where the atoms of the transform dictionary are generalized
        translations of each graph spectral filter to each vertex on the graph:

        .. math:: c = D^* s,

        where the columns of :math:`D` are :math:`g_{i,m} = T_i g_m` and
        :math:`T_i` is a generalized translation operator applied to each
        filter :math:`\hat{g}_m(\cdot)`. Each column of :math:`c` is the
        response of the signal to one filter.

        In other words, this function is applying the analysis operator
        :math:`D^*`, respectively the synthesis operator :math:`D`, associated
        with the frame defined by the filter bank to the signals.

        Parameters
        ----------
        s : array_like
            Graph signals, a tensor of shape ``(N_NODES, N_SIGNALS,
            N_FEATURES)``, where ``N_NODES`` is the number of nodes in the
            graph, ``N_SIGNALS`` the number of independent signals you want to
            filter, and ``N_FEATURES`` is either 1 (analysis) or the number of
            filters in the filter bank (synthesis).
        method : {'exact', 'chebyshev'}
            Whether to use the exact method (via the graph Fourier transform)
            or the Chebyshev polynomial approximation. A Lanczos
            approximation is coming.
        order : int
            Degree of the Chebyshev polynomials.

        Returns
        -------
        s : ndarray
            Graph signals, a tensor of shape ``(N_NODES, N_SIGNALS,
            N_FEATURES)``, where ``N_NODES`` and ``N_SIGNALS`` are the number
            of nodes and signals of the signal tensor that pas passed in, and
            ``N_FEATURES`` is either 1 (synthesis) or the number of filters in
            the filter bank (analysis).

        References
        ----------
        See :cite:`hammond2011wavelets` for details on filtering graph signals.

        Examples
        --------

        Create a bunch of smooth signals by low-pass filtering white noise:

        >>> import matplotlib.pyplot as plt
        >>> G = graphs.Ring(N=60)
        >>> G.estimate_lmax()
        >>> s = np.random.default_rng(42).uniform(size=(G.N, 10))
        >>> taus = [1, 10, 100]
        >>> s = filters.Heat(G, taus).filter(s)
        >>> s.shape
        (60, 10, 3)

        Plot the 3 smoothed versions of the 10th signal:

        >>> fig, ax = plt.subplots()
        >>> G.set_coordinates('line1D')  # To visualize multiple signals in 1D.
        >>> _ = G.plot(s[:, 9, :], ax=ax)
        >>> legend = [r'$\tau={}$'.format(t) for t in taus]
        >>> ax.legend(legend)  # doctest: +ELLIPSIS
        <matplotlib.legend.Legend object at ...>

        Low-pass filter a delta to create a localized smooth signal:

        >>> G = graphs.Sensor(30, seed=42)
        >>> G.compute_fourier_basis()  # Reproducible computation of lmax.
        >>> s1 = np.zeros(G.N)
        >>> s1[13] = 1
        >>> s1 = filters.Heat(G, 3).filter(s1)
        >>> s1.shape
        (30,)

        Filter and reconstruct our signal:

        >>> g = filters.MexicanHat(G, Nf=4)
        >>> s2 = g.analyze(s1)
        >>> s2.shape
        (30, 4)
        >>> s2 = g.synthesize(s2)
        >>> s2.shape
        (30,)

        Look how well we were able to reconstruct:

        >>> fig, axes = plt.subplots(1, 2)
        >>> _ = G.plot(s1, ax=axes[0])
        >>> _ = G.plot(s2, ax=axes[1])
        >>> print('{:.5f}'.format(np.linalg.norm(s1 - s2)))
        0.27649

        Perfect reconstruction with Itersine, a tight frame:

        >>> g = filters.Itersine(G)
        >>> s2 = g.analyze(s1, method='exact')
        >>> s2 = g.synthesize(s2, method='exact')
        >>> np.linalg.norm(s1 - s2) < 1e-10
        True

        """
        s = self.G._check_signal(s)

        # TODO: not in self.Nin (Nf = Nin x Nout).
        if s.ndim == 1 or s.shape[-1] not in [1, self.Nf]:
            if s.ndim == 3:
                raise ValueError('Third dimension (#features) should be '
                                 'either 1 or the number of filters Nf = {}, '
                                 'got {}.'.format(self.Nf, s.shape))
            s = np.expand_dims(s, -1)
        n_features_in = s.shape[-1]

        if s.ndim < 3:
            s = np.expand_dims(s, 1)
        n_signals = s.shape[1]

        if s.ndim > 3:
            raise ValueError('At most 3 dimensions: '
                             '#nodes x #signals x #features.')
        assert s.ndim == 3

        # TODO: generalize to 2D (m --> n) filter banks.
        # Only 1 --> Nf (analysis) and Nf --> 1 (synthesis) for now.
        n_features_out = self.Nf if n_features_in == 1 else 1

        if method == 'exact':

            # TODO: will be handled by g.adjoint().
            axis = 1 if n_features_in == 1 else 2
            f = self.evaluate(self.G.e)
            f = np.expand_dims(f.T, axis)
            assert f.shape == (self.G.N, n_features_in, n_features_out)

            s = self.G.gft(s)
            s = np.matmul(s, f)
            s = self.G.igft(s)

        elif method == 'chebyshev':

            # TODO: update Chebyshev implementation (after 2D filter banks).
            c = approximations.compute_cheby_coeff(self, m=order)

            if n_features_in == 1:  # Analysis.
                s = s.squeeze(axis=2)
                s = approximations.cheby_op(self.G, c, s)
                s = s.reshape((self.G.N, n_features_out, n_signals), order='F')
                s = s.swapaxes(1, 2)

            elif n_features_in == self.Nf:  # Synthesis.
                s = s.swapaxes(1, 2)
                s_in = s.reshape(
                    (self.G.N * n_features_in, n_signals), order='F')
                s = np.zeros((self.G.N, n_signals))
                tmpN = np.arange(self.G.N, dtype=int)
                for i in range(n_features_in):
                    s += approximations.cheby_op(self.G,
                                                 c[i],
                                                 s_in[i * self.G.N + tmpN])
                s = np.expand_dims(s, 2)

        else:
            raise ValueError('Unknown method {}.'.format(method))

        # Return a 1D signal if e.g. a 1D signal was filtered by one filter.
        return s.squeeze()

    def analyze(self, s, method='chebyshev', order=30):
        r"""Convenience alias to :meth:`filter`."""
        if s.ndim == 3 and s.shape[-1] != 1:
            raise ValueError('Last dimension (#features) should be '
                             '1, got {}.'.format(s.shape))
        return self.filter(s, method, order)

    def synthesize(self, s, method='chebyshev', order=30):
        r"""Convenience wrapper around :meth:`filter`.

        Will be an alias to `adjoint().filter()` in the future.
        """
        if s.shape[-1] != self.Nf:
            raise ValueError('Last dimension (#features) should be the number '
                             'of filters Nf = {}, got {}.'.format(self.Nf,
                                                                  s.shape))
        return self.filter(s, method, order)

    def localize(self, i, **kwargs):
        r"""Localize the kernels at a node (to visualize them).

        That is particularly useful to visualize a filter in the vertex domain.

        A kernel is localized on vertex :math:`v_i` by filtering a Kronecker
        delta :math:`\delta_i` as

        .. math:: (g(L) \delta_i)(j) = g(L)(i,j),
                  \text{ where } \delta_i(j) =
                  \begin{cases} 0 \text{ if } i \neq j, \\
                                1 \text{ if } i = j.    \end{cases}

        Parameters
        ----------
        i : int
            Index of the node where to localize the kernel.
        kwargs: dict
            Parameters to be passed to the :meth:`analyze` method.

        Returns
        -------
        s : ndarray
            Kernel localized at vertex i.

        Examples
        --------
        Visualize heat diffusion on a grid by localizing the heat kernel.

        >>> import matplotlib
        >>> N = 20
        >>> DELTA = N//2 * (N+1)
        >>> G = graphs.Grid2d(N)
        >>> G.estimate_lmax()
        >>> g = filters.Heat(G, 100)
        >>> s = g.localize(DELTA)
        >>> _ = G.plot(s, highlight=DELTA)

        """
        s = np.zeros(self.G.N)
        s[i] = 1
        return np.sqrt(self.G.N) * self.filter(s, **kwargs)

    def estimate_frame_bounds(self, x=None):
        r"""Estimate lower and upper frame bounds.

        A filter bank forms a frame if there are positive real numbers
        :math:`A` and :math:`B`, :math:`0 < A \leq B < \infty`, that satisfy
        the *frame condition*

        .. math:: A \|x\|^2 \leq \| g(L) x \|^2 \leq B \|x\|^2

        for all signals :math:`x \in \mathbb{R}^N`, where :math:`g(L)` is the
        analysis operator of the filter bank.

        As :math:`g(L) = U g(\Lambda) U^\top` is diagonalized by the Fourier
        basis :math:`U` with eigenvalues :math:`\Lambda`, :math:`\| g(L) x \|^2
        = \| g(\Lambda) U^\top x \|^2`, and :math:`A = \min g^2(\Lambda)`,
        :math:`B = \max g^2(\Lambda)`.

        Parameters
        ----------
        x : array_like
            Graph frequencies at which to evaluate the filter bank `g(x)`.
            The default is ``x = np.linspace(0, G.lmax, 1000)``.
            The exact bounds are given by evaluating the filter bank at the
            eigenvalues of the graph Laplacian, i.e., ``x = G.e``.

        Returns
        -------
        A : float
            Lower frame bound of the filter bank.
        B : float
            Upper frame bound of the filter bank.

        See Also
        --------
        compute_frame: compute the frame
        complement: complement a filter bank to become a tight frame

        Examples
        --------

        >>> from matplotlib import pyplot as plt
        >>> fig, axes = plt.subplots(2, 2, figsize=(10, 6))
        >>> G = graphs.Sensor(64, seed=42)
        >>> G.compute_fourier_basis()
        >>> g = filters.Abspline(G, 7)

        Estimation quality vs speed (loose & fast -> exact & slow):

        >>> A, B = g.estimate_frame_bounds(np.linspace(0, G.lmax, 5))
        >>> print('A={:.3f}, B={:.3f}'.format(A, B))
        A=1.883, B=2.288
        >>> A, B = g.estimate_frame_bounds()
        >>> print('A={:.3f}, B={:.3f}'.format(A, B))
        A=1.708, B=2.359
        >>> A, B = g.estimate_frame_bounds(G.e)
        >>> print('A={:.3f}, B={:.3f}'.format(A, B))
        A=1.875, B=2.359

        The frame bounds can be seen in the plot of the filter bank as the
        minimum and maximum of their squared sum (the black curve):

        >>> def plot(g, ax):
        ...     g.plot(ax=ax, labels=False, title='')
        ...     ax.hlines(B, 0, G.lmax, colors='r', zorder=3,
        ...               label='upper bound $B={:.2f}$'.format(B))
        ...     ax.hlines(A, 0, G.lmax, colors='b', zorder=3,
        ...               label='lower bound $A={:.2f}$'.format(A))
        ...     ax.legend(loc='center right')
        >>> plot(g, axes[0, 0])

        The heat kernel has a null-space and doesn't define a frame (the lower
        bound should be greater than 0 to have a frame):

        >>> g = filters.Heat(G)
        >>> A, B = g.estimate_frame_bounds()
        >>> print('A={:.3f}, B={:.3f}'.format(A, B))
        A=0.000, B=1.000
        >>> plot(g, axes[0, 1])

        Without a null-space, the heat kernel forms a frame:

        >>> g = filters.Heat(G, scale=[1, 10])
        >>> A, B = g.estimate_frame_bounds()
        >>> print('A={:.3f}, B={:.3f}'.format(A, B))
        A=0.135, B=2.000
        >>> plot(g, axes[1, 0])

        A kernel and its dual form a tight frame (A=B):

        >>> g = filters.Regular(G)
        >>> A, B = g.estimate_frame_bounds()
        >>> print('A={:.3f}, B={:.3f}'.format(A, B))
        A=1.000, B=1.000
        >>> plot(g, axes[1, 1])
        >>> fig.tight_layout()

        The Itersine filter bank forms a tight frame (A=B):

        >>> g = filters.Itersine(G)
        >>> A, B = g.estimate_frame_bounds()
        >>> print('A={:.3f}, B={:.3f}'.format(A, B))
        A=1.000, B=1.000

        """
        if x is None:
            x = np.linspace(0, self.G.lmax, 1000)
        else:
            x = np.asanyarray(x)

        sum_filters = np.sum(self.evaluate(x)**2, axis=0)

        return sum_filters.min(), sum_filters.max()

    def compute_frame(self, **kwargs):
        r"""Compute the associated frame.

        A filter bank defines a frame, which is a generalization of a basis to
        sets of vectors that may be linearly dependent. See
        `Wikipedia <https://en.wikipedia.org/wiki/Frame_(linear_algebra)>`_.

        The frame of a filter bank is the union of the frames of its
        constituent filters. The vectors forming the frame are the rows of the
        *analysis operator*

        .. math::
            g(L) = \begin{pmatrix} g_1(L) \\ \vdots \\ g_F(L) \end{pmatrix}
                   \in \mathbb{R}^{NF \times N}, \quad
            g_i(L) = U g_i(\Lambda) U^\top,

        where :math:`g_i` are the filter kernels, :math:`N` is the number of
        nodes, :math:`F` is the number of filters, :math:`L` is the graph
        Laplacian, :math:`\Lambda` is a diagonal matrix of the Laplacian's
        eigenvalues, and :math:`U` is the Fourier basis, i.e., its columns are
        the eigenvectors of the Laplacian.

        The matrix :math:`g(L)` represents the *analysis operator* of the
        frame. Its adjoint :math:`g(L)^\top` represents the *synthesis
        operator*. A signal :math:`x` is thus analyzed with the frame by
        :math:`y = g(L) x`, and synthesized from its frame coefficients by
        :math:`z = g(L)^\top y`. Computing this matrix is however a rather
        inefficient way of doing those operations.

        If :math:`F > 1`, the frame is said to be over-complete and the
        representation :math:`g(L) x` of the signal :math:`x` is said to be
        redundant.

        If the frame is tight, the *frame operator* :math:`g(L)^\top g(L)` is
        diagonal, with entries equal to the frame bound :math:`A = B`.

        Parameters
        ----------
        kwargs: dict
            Parameters to be passed to the :meth:`analyze` method.

        Returns
        -------
        frame : ndarray
            Array of size (#nodes x #filters) x #nodes.

        See Also
        --------
        estimate_frame_bounds: estimate the frame bounds
        filter: more efficient way to filter signals

        Examples
        --------

        >>> G = graphs.Sensor(100, seed=42)
        >>> G.compute_fourier_basis()

        Filtering as a multiplication with the matrix representation of the
        frame analysis operator:

        >>> g = filters.MexicanHat(G, Nf=6)
        >>> s = np.random.default_rng().uniform(size=G.N)
        >>>
        >>> gL = g.compute_frame()
        >>> gL.shape
        (600, 100)
        >>> s1 = gL.dot(s)
        >>> s1 = s1.reshape(G.N, -1, order='F')
        >>>
        >>> s2 = g.filter(s)
        >>> np.all(np.abs(s1 - s2) < 1e-10)
        True

        The frame operator of a tight frame is the identity matrix times the
        frame bound:

        >>> g = filters.Itersine(G)
        >>> A, B = g.estimate_frame_bounds()
        >>> print('A={:.3f}, B={:.3f}'.format(A, B))
        A=1.000, B=1.000
        >>> gL = g.compute_frame(method='exact')
        >>> gL.shape
        (600, 100)
        >>> np.all(gL.T.dot(gL) - np.identity(G.N) < 1e-10)
        True

        """
        if self.G.N > 2000:
            _logger.warning('Creating a big matrix. '
                            'You should prefer the filter method.')

        # Filter one delta per vertex.
        s = np.identity(self.G.N)
        return self.filter(s, **kwargs).T.reshape(-1, self.G.N)

    def complement(self, frame_bound=None):
        r"""Return the filter that makes the frame tight.

        The complementary filter is designed such that the union of a filter
        bank and its complementary filter forms a tight frame.

        Parameters
        ----------
        frame_bound : float or None
            The desired frame bound :math:`A = B` of the resulting tight frame.
            The chosen bound should be larger than the sum of squared
            evaluations of all filters in the filter bank. If None (the
            default), the method chooses the smallest feasible bound.

        Returns
        -------
        complement: Filter
            The complementary filter.

        See Also
        --------
        estimate_frame_bounds: estimate the frame bounds

        Examples
        --------
        >>> from matplotlib import pyplot as plt
        >>> G = graphs.Sensor(100, seed=42)
        >>> G.estimate_lmax()
        >>> g = filters.Abspline(G, 4)
        >>> A, B = g.estimate_frame_bounds()
        >>> print('A={:.3f}, B={:.3f}'.format(A, B))
        A=0.200, B=1.971
        >>> fig, axes = plt.subplots(1, 2)
        >>> fig, ax = g.plot(ax=axes[0])
        >>> g += g.complement()
        >>> A, B = g.estimate_frame_bounds()
        >>> print('A={:.3f}, B={:.3f}'.format(A, B))
        A=1.971, B=1.971
        >>> fig, ax = g.plot(ax=axes[1])

        """

        def kernel(x, *args, **kwargs):

            y = self.evaluate(x)
            np.power(y, 2, out=y)
            y = np.sum(y, axis=0)

            if frame_bound is None:
                bound = y.max()
            elif y.max() > frame_bound:
                raise ValueError('The chosen bound is not feasible. '
                                 'Choose at least {}.'.format(y.max()))
            else:
                bound = frame_bound

            return np.sqrt(bound - y)

        return Filter(self.G, kernel)

    def inverse(self):
        r"""Return the pseudo-inverse filter bank.

        The pseudo-inverse of the *analysis filter bank* :math:`g` is the
        *synthesis filter bank* :math:`g^+` such that

        .. math:: g(L)^+ g(L) = I,

        where :math:`I` is the identity matrix, and the *synthesis operator*

        .. math:: g(L)^+ = (g(L)\top g(L))^{-1} g(L)^\top
                         = (g_1(L)^+, \dots, g_F(L)^+)
                           \in \mathbb{R}^{N \times NF}

        is the left pseudo-inverse of the analysis operator :math:`g(L)`. Note
        that :math:`g_i(L)^+` is the pseudo-inverse of :math:`g_i(L)`,
        :math:`N` is the number of vertices, and :math:`F` is the number of
        filters in the bank.

        The above relation holds, and the reconstruction is exact, if and only
        if :math:`g(L)` is a frame. To be a frame, the rows of :math:`g(L)`
        must span the whole space (i.e., :math:`g(L)` must have full row rank).
        That is the case if the lower frame bound :math:`A > 0`. If
        :math:`g(L)` is not a frame, the reconstruction :math:`g(L)^+ g(L) x`
        will be the closest to :math:`x` in the least square sense.

        While there exists infinitely many inverses of the analysis operator of
        a frame, the pseudo-inverse is unique and corresponds to the *canonical
        dual* of the filter kernel.

        The *frame operator* of :math:`g^+` is :math:`g(L)^+ (g(L)^+)^\top =
        (g(L)\top g(L))^{-1}`, the inverse of the frame operator of :math:`g`.
        Similarly, its *frame bounds* are :math:`A^{-1}` and :math:`B^{-1}`,
        where :math:`A` and :math:`B` are the frame bounds of :math:`g`.

        If :math:`g` is tight (i.e., :math:`A=B`), the canonical dual is given
        by :math:`g^+ = g / A` (i.e., :math:`g^+_i = g_i / A \ \forall i`).

        Returns
        -------
        inverse : :class:`pygsp2.filters.Filter`
            The pseudo-inverse filter bank, which synthesizes (or reconstructs)
            a signal from its coefficients using the canonical dual frame.

        See Also
        --------
        estimate_frame_bounds: estimate the frame bounds

        Examples
        --------
        >>> from matplotlib import pyplot as plt
        >>> G = graphs.Sensor(100, seed=42)
        >>> G.compute_fourier_basis()
        >>> # Create a filter and its inverse.
        >>> g = filters.Abspline(G, 5)
        >>> h = g.inverse()
        >>> # Plot them.
        >>> fig, axes = plt.subplots(1, 2)
        >>> _ = g.plot(ax=axes[0], title='original filter bank')
        >>> _ = h.plot(ax=axes[1], title='inverse filter bank')
        >>> # Filtering with the inverse reconstructs the original signal.
        >>> x = np.random.default_rng(42).normal(size=G.N)
        >>> y = g.filter(x, method='exact')
        >>> z = h.filter(y, method='exact')
        >>> np.linalg.norm(x - z) < 1e-10
        True
        >>> # Indeed, they cancel each others' effect.
        >>> Ag, Bg = g.estimate_frame_bounds()
        >>> Ah, Bh = h.estimate_frame_bounds()
        >>> print('A(g)*B(h) = {:.3f} * {:.3f} = {:.3f}'.format(Ag, Bh, Ag*Bh))
        A(g)*B(h) = 0.687 * 1.457 = 1.000
        >>> print('B(g)*A(h) = {:.3f} * {:.3f} = {:.3f}'.format(Bg, Ah, Bg*Ah))
        B(g)*A(h) = 1.994 * 0.501 = 1.000

        """

        A, _ = self.estimate_frame_bounds()
        if A == 0:
            _logger.warning('The filter bank is not invertible as it is not '
                            'a frame (lower frame bound A=0).')

        def kernel(g, i, x):
            y = g.evaluate(x).T
            z = np.linalg.pinv(np.expand_dims(y, axis=-1)).squeeze(axis=-2)
            return z[:, i]  # Return one filter.

        kernels = [partial(kernel, self, i) for i in range(self.n_filters)]

        return Filter(self.G, kernels)

    def plot(self, n=500, eigenvalues=None, sum=None, labels=None, title=None,
             ax=None, **kwargs):
        r"""Docstring overloaded at import time."""
        from pygsp2.plotting import _plot_filter
        return _plot_filter(self, n=n, eigenvalues=eigenvalues, sum=sum,
                            labels=labels, title=title, ax=ax, **kwargs)



# -*- coding: utf-8 -*-

from pygsp2 import utils
from . import Filter  # prevent circular import in Python < 3.5


class Gabor(Filter):
    r"""Design a filter bank with a kernel centered at each frequency.

    Design a filter bank from translated versions of a mother filter.
    The mother filter is translated to each eigenvalue of the Laplacian.
    That is equivalent to convolutions with deltas placed at those eigenvalues.

    In classical image processing, a Gabor filter is a sinusoidal wave
    multiplied by a Gaussian function (here, the kernel). It analyzes whether
    there are any specific frequency content in the image in specific
    directions in a localized region around the point of analysis. This
    implementation for graph signals allows arbitrary (but isotropic) kernels.

    This filter bank can be used to compute the frequency content of a signal
    at each vertex. After filtering, one obtains a vertex-frequency
    representation :math:`Sf(i,k)` of a signal :math:`f` as

    .. math:: Sf(i, k) = \langle g_{i,k}, f \rangle,

    where :math:`g_{i,k}` is the mother kernel centered on eigenvalue
    :math:`\lambda_k` and localized on vertex :math:`v_i`.

    While :math:`g_{i,k}` should ideally be localized in both the spectral and
    vertex domains, that is impossible for some graphs due to the localization
    of some eigenvectors. See :attr:`pygsp2.graphs.Graph.coherence`.

    Parameters
    ----------
    graph : :class:`pygsp2.graphs.Graph`
    kernel : :class:`pygsp2.filters.Filter`
        Kernel function to be centered at each graph frequency (eigenvalue of
        the graph Laplacian).

    See Also
    --------
    Modulation : Another way to translate a filter in the spectral domain.

    Notes
    -----
    The eigenvalues of the graph Laplacian (i.e., the Fourier basis) are needed
    to center the kernels.

    Examples
    --------

    Filter bank's representation in Fourier and time (path graph) domains.

    >>> import matplotlib.pyplot as plt
    >>> G = graphs.Path(N=7)
    >>> G.compute_fourier_basis()
    >>> G.set_coordinates('line1D')
    >>>
    >>> g1 = filters.Expwin(G, band_min=None, band_max=0, slope=3)
    >>> g2 = filters.Rectangular(G, band_min=-0.05, band_max=0.05)
    >>> g3 = filters.Heat(G, scale=10)
    >>>
    >>> fig, axes = plt.subplots(3, 2, figsize=(10, 10))
    >>> for g, ax in zip([g1, g2, g3], axes):
    ...     g = filters.Gabor(G, g)
    ...     s = g.localize(G.N // 2, method='exact')
    ...     _ = g.plot(ax=ax[0], sum=False)
    ...     _ = G.plot(s, ax=ax[1])
    >>> fig.tight_layout()

    """

    def __init__(self, graph, kernel):

        if kernel.n_filters != 1:
            raise ValueError('A kernel must be one filter. The passed '
                             'filter bank {} has {}.'.format(
                                 kernel, kernel.n_filters))
        if kernel.G is not graph:
            raise ValueError('The graph passed to this filter bank must '
                             'be the one used to build the mother kernel.')

        kernels = []
        for i in range(graph.n_vertices):
            kernels.append(lambda x, i=i: kernel.evaluate(x - graph.e[i]))

        super(Gabor, self).__init__(graph, kernels)

    def filter(self, s, method='exact', order=None):
        """TODO: indirection will be removed when poly filtering is merged."""
        return super(Gabor, self).filter(s, method='exact')



r"""
Fourier transform
=================

The graph Fourier transform :meth:`pygsp2.graphs.Graph.gft` transforms a
signal from the vertex domain to the spectral domain. The smoother the signal
(see :meth:`pygsp2.graphs.Graph.dirichlet_energy`), the lower in the frequencies
its energy is concentrated.
"""

import numpy as np
from matplotlib import pyplot as plt
import pygsp2 as pg

G = pg.graphs.Sensor(seed=42)
G.compute_fourier_basis()

scales = [10, 3, 0]
limit = 0.5

fig, axes = plt.subplots(2, len(scales), figsize=(12, 4))
fig.subplots_adjust(hspace=0.5)

x0 = np.random.default_rng(1).normal(size=G.N)
for i, scale in enumerate(scales):
    g = pg.filters.Heat(G, scale)
    x = g.filter(x0).squeeze()
    x /= np.linalg.norm(x)
    x_hat = G.gft(x).squeeze()
    
    G.plot(x, limits=[-limit, limit], ax=axes[0, i])
    axes[0, i].set_axis_off()
    axes[0, i].set_title('$x^T L x = {:.2f}$'.format(G.dirichlet_energy(x)))

    axes[1, i].plot(G.e, np.abs(x_hat), '.-')
    axes[1, i].set_xticks(range(0, 16, 4))
    axes[1, i].set_xlabel(r'graph frequency $\lambda$')
    axes[1, i].set_ylim(-0.05, 0.95)

axes[1, 0].set_ylabel(r'frequency content $\hat{x}(\lambda)$')

# axes[0, 0].set_title(r'$x$: signal in the vertex domain')
# axes[1, 0].set_title(r'$\hat{x}$: signal in the spectral domain')

fig.tight_layout()
plt.show()


# -*- coding: utf-8 -*-

from pygsp2 import utils
from pygsp2.graphs import NNGraph  # prevent circular import in Python < 3.5


class Bunny(NNGraph):
    r"""Stanford bunny (NN-graph).

    References
    ----------
    See :cite:`turk1994zippered`.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> G = graphs.Bunny()
    >>> fig = plt.figure()
    >>> ax1 = fig.add_subplot(121)
    >>> ax2 = fig.add_subplot(122, projection='3d')
    >>> _ = ax1.spy(G.W, markersize=0.1)
    >>> _ = _ = G.plot(ax=ax2)

    """

    def __init__(self, **kwargs):

        data = utils.loadmat('pointclouds/bunny')

        plotting = {
            'vertex_size': 10,
            'elevation': -90,
            'azimuth': 90,
            'distance': 8,
        }

        super(Bunny, self).__init__(Xin=data['bunny'],
                                    epsilon=0.02, NNtype='radius',
                                    center=False, rescale=False,
                                    plotting=plotting, **kwargs)



# -*- coding: utf-8 -*-

import numpy as np
from scipy import sparse

from . import Graph  # prevent circular import in Python < 3.5


class Torus(Graph):
    r"""Sampled torus manifold.

    On the torus, the graph Fourier transform (GFT) is the Kronecker product
    between the GFT of two :class:`~pygsp2.graphs.Ring` graphs.

    Parameters
    ----------
    Nv : int
        Number of vertices along the first dimension.
    Mv : int
        Number of vertices along the second dimension. Default is ``Nv``.

    See Also
    --------
    Ring : 1D line with periodic boundary conditions
    Grid2d : Kronecker product of two path graphs

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> G = graphs.Torus(10)
    >>> fig = plt.figure()
    >>> ax1 = fig.add_subplot(121)
    >>> ax2 = fig.add_subplot(122, projection='3d')
    >>> _ = ax1.spy(G.W, markersize=1.5)
    >>> _ = G.plot(ax=ax2)
    >>> _ = ax2.set_zlim(-1.5, 1.5)

    """

    def __init__(self, Nv=16, Mv=None, **kwargs):

        if Mv is None:
            Mv = Nv

        self.Nv = Nv
        self.Mv = Mv

        # Create weighted adjancency matrix
        K = 2 * Nv
        J = 2 * Mv
        i_inds = np.zeros((K*Mv + J*Nv), dtype=float)
        j_inds = np.zeros((K*Mv + J*Nv), dtype=float)

        tmpK = np.arange(K, dtype=int)
        tmpNv1 = np.arange(Nv - 1)
        tmpNv = np.arange(Nv)

        for i in range(Mv):
            i_inds[i*K + tmpK] = i*Nv + \
                np.concatenate((np.array([Nv - 1]), tmpNv1, tmpNv))

            j_inds[i*K + tmpK] = i*Nv + \
                np.concatenate((tmpNv, np.array([Nv - 1]), tmpNv1))

        tmp2Nv = np.arange(2*Nv, dtype=int)

        for i in range(Mv - 1):
            i_inds[K*Mv + i*2*Nv + tmp2Nv] = \
                np.concatenate((i*Nv + tmpNv, (i + 1)*Nv + tmpNv))

            j_inds[K*Mv + i*2*Nv + tmp2Nv] = \
                np.concatenate(((i + 1)*Nv + tmpNv, i*Nv + tmpNv))

        i_inds[K*Mv + (Mv - 1)*2*Nv + tmp2Nv] = \
            np.concatenate((tmpNv, (Mv - 1)*Nv + tmpNv))

        j_inds[K*Mv + (Mv - 1)*2*Nv + tmp2Nv] = \
            np.concatenate(((Mv - 1)*Nv + tmpNv, tmpNv))

        W = sparse.csc_matrix((np.ones((K*Mv + J*Nv)), (i_inds, j_inds)),
                              shape=(Mv*Nv, Mv*Nv))

        # Create coordinate
        T = 1.5 + np.sin(np.arange(Mv)*2*np.pi/Mv).reshape(1, Mv)
        U = np.cos(np.arange(Mv)*2*np.pi/Mv).reshape(1, Mv)
        xtmp = np.cos(np.arange(Nv).reshape(Nv, 1)*2*np.pi/Nv)*T
        ytmp = np.sin(np.arange(Nv).reshape(Nv, 1)*2*np.pi/Nv)*T
        ztmp = np.kron(np.ones((Nv, 1)), U)
        coords = np.concatenate((np.reshape(xtmp, (Mv*Nv, 1), order='F'),
                                 np.reshape(ytmp, (Mv*Nv, 1), order='F'),
                                 np.reshape(ztmp, (Mv*Nv, 1), order='F')),
                                axis=1)

        plotting = {
            'vertex_size': 60,
            'limits': np.array([-2.5, 2.5, -2.5, 2.5, -2.5, 2.5])
        }

        super(Torus, self).__init__(W, coords=coords,
                                    plotting=plotting, **kwargs)

    def _get_extra_repr(self):
        return dict(Nv=self.Nv, Mv=self.Mv)
